/**
    @file
   @author Shin'ichiro Nakaoka
*/

#include "OSGSceneItem.h"
#include "ItemTreeView.h"
#include "ItemManager.h"
#include "Archive.h"
#include "RootItem.h"
#include "OptionManager.h"
#include "PutPropertyFunction.h"
#include <cnoid/YAMLReader>
#include <cnoid/FileUtil>
#include <boost/bind.hpp>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

namespace {

    const bool TRACE_FUNCTIONS = false;

    bool loadOSGSceneItem(OSGSceneItem* item, const std::string& filename, std::ostream& os)
    {
        OSGSceneObject::ReadResult result = item->loadScene(filename);
        if(result.success()){
            if(item->name().empty()){
                item->setName(getFilename(filesystem::path(filename)));
            }
        } else {
            os << result.message();
        }

        return result.success();
    }


    class OSGSceneItemManager
    {
        std::set<OSGSceneItemPtr> visibleOSGSceneItems;

    public:

        OSGSceneItemManager(ExtensionManager* ext) {

            ext->itemManager().registerClass<OSGSceneItem>("OSGSceneItem");

            ext->itemManager().addLoader<OSGSceneItem>(
                _("Scene File (OSG)"), "OSG-IMPORTABLE-SCENE-FILE", "",
                bind(loadOSGSceneItem, _1, _2, _3), ItemManager::PRIORITY_CONVERSION);

            ext->optionManager().addOption("model", program_options::value< vector<string> >(), "load a 3d model file");
            ext->optionManager().sigOptionsParsed().connect(
                bind(&OSGSceneItemManager::onSigOptionsParsed, this, _1));

            ItemTreeView::mainInstance()->sigCheckToggled().connect(
                bind(&OSGSceneItemManager::onItemCheckToggled, this, _1, _2));
        }

        void onSigOptionsParsed(boost::program_options::variables_map& variables) {

            if(variables.count("model")){
                vector<string> files = variables["model"].as< vector<string> >();
                for(size_t i=0; i < files.size(); ++i){
                    OSGSceneItemPtr item(new OSGSceneItem());
                    if(item->load(files[i], "OSG-IMPORTABLE-SCENE-FILE")){
                        RootItem::mainInstance()->addChildItem(item);
                        ItemTreeView::mainInstance()->checkItem(item, true);
                    }
                }
            }
        }

        void onItemCheckToggled(Item* item, bool isChecked) {
            OSGSceneItem* sceneItem = dynamic_cast<OSGSceneItem*>(item);
            if(sceneItem){
                std::set<OSGSceneItemPtr>::iterator p = visibleOSGSceneItems.find(sceneItem);
                if(isChecked){
                    if(p == visibleOSGSceneItems.end()){
                        sceneItem->sceneObject()->show();
                        visibleOSGSceneItems.insert(sceneItem);
                    }
                } else {
                    if(p != visibleOSGSceneItems.end()){
                        sceneItem->sceneObject()->hide();
                        visibleOSGSceneItems.erase(p);
                    }
                }
            }
        }

        ~OSGSceneItemManager() {
            std::set<OSGSceneItemPtr>::iterator p;
            for(p = visibleOSGSceneItems.begin(); p != visibleOSGSceneItems.end(); ++p){
                OSGSceneItemPtr sceneItem = *p;
                sceneItem->sceneObject()->hide();
            }
        }
    };
}
    

void OSGSceneItem::initialize(ExtensionManager* ext)
{
    static bool initialized = false;
    if(!initialized){
        ext->manage(new OSGSceneItemManager(ext));
        initialized = true;
    }
}


OSGSceneItem::OSGSceneItem()
    : sceneObject_(new OSGSceneObject)
{
    setAttribute(LOAD_ONLY);
}


OSGSceneItem::OSGSceneItem(const OSGSceneItem& org)
    : Item(org),
      sceneObject_(new OSGSceneObject(*org.sceneObject_))
{

}


OSGSceneItem::~OSGSceneItem()
{

}


void OSGSceneItem::onDisconnectedFromRoot()
{
    sceneObject_->hide();
}


OSGSceneObject::ReadResult OSGSceneItem::loadScene(const std::string& filename)
{
    return sceneObject_->load(filename);
}


ItemPtr OSGSceneItem::doDuplicate() const
{
    return new OSGSceneItem(*this);
}


void OSGSceneItem::doPutProperties(PutPropertyFunction& putProperty)
{
    putProperty(_("Scene file"), getFilename(filesystem::path(lastAccessedFilePath())));
}


bool OSGSceneItem::store(Archive& archive)
{
    if(!lastAccessedFilePath().empty()){
        archive.writeRelocatablePath("file", lastAccessedFilePath());
        archive.write("format", lastAccessedFileFormatId());
    }
    return true;
}


bool OSGSceneItem::restore(const Archive& archive)
{
    std::string filename, formatId;
    if(archive.readRelocatablePath("file", filename) && archive.read("format", formatId)){
        if(load(filename, archive.currentParentItem(), formatId)){
            return true;
        }
    }
    return false;
}
