/**
   @author Shin'ichiro Nakaoka
*/

#ifndef CNOID_BASE_OSG_SCENE_OBJECT_H_INCLUDED
#define CNOID_BASE_OSG_SCENE_OBJECT_H_INCLUDED

#include <boost/signal.hpp>
#include <osg/Group>
#include <osg/Camera>
#include <osgDB/ReaderWriter>
#include <osgGA/GUIEventAdapter>
#include "exportdecl.h"

namespace cnoid {

    class MenuManager;
    class OSGSceneView;
    class OSGSceneViewImpl;
    class Item;

    class CNOID_EXPORT OSGSceneViewEvent
    {
      public:

        OSGSceneViewEvent(const OSGSceneViewEvent& org);
        
        inline const osg::Vec3d& point() const { return point_; }
        inline const osg::Vec3d& normal() const { return normal_; }
        inline const osg::NodePath& path() const { return path_; }
        inline osg::Camera* camera() const { return camera_.get(); }

        inline double x() const { return x_; }

        /**
           The value is increasing from the bottom to top (the coordinate of the OpenGL viewport).
        */
        inline double y() const { return y_; }

        inline int px() const { return px_; }
        inline int py() const { return py_; }

        /**
           @return a osgGA::GUIEventAdapter::KeySymbol value
        */
        inline int key() const { return key_; }

        /**
           @return a osgGA::GUIEventAdapter::MouseButtonMask value
        */
        inline int button() const { return button_; }

        /**
           @return a osgGA::GUIEventAdapter::ModKeyMask value
        */
        inline int modKeyMask() const { return modKeyMask_; }
        
        /**
           @return a osgGA::GUIEventAdapter::ScrollingMotion value
        */
        inline int scrollingMotion() const { return scrollingMotion_; }

        inline double scrollingDelta() const { return scrollingDelta_; }
        
        void updateIndicator(const std::string& message) const;

      private:

        osg::ref_ptr<osg::Camera> camera_;
        osg::Vec3d point_;
        osg::Vec3d normal_;
        osg::NodePath path_;
        double x_;
        double y_;
        int px_;
        int py_;
        int key_;
        int modKeyMask_;
        int button_;
        int scrollingMotion_;
        double scrollingDelta_;
        OSGSceneView* sceneView;

        OSGSceneViewEvent();
        OSGSceneViewEvent& operator=(const OSGSceneViewEvent& org); // disabled

        friend class OSGSceneViewImpl;
    };
    

    class CNOID_EXPORT OSGSceneObject : public osg::Group
    {
      public:
        OSGSceneObject();
        OSGSceneObject(const OSGSceneObject& org, const osg::CopyOp& copyop = osg::CopyOp::SHALLOW_COPY);

        virtual const char* libraryName() const;
        virtual const char* className () const;

        typedef osgDB::ReaderWriter::ReadResult ReadResult;

        ReadResult load(const std::string filename);

        void show(bool on = true);
        void hide();

        inline bool isActive() { return isActive_; }

        // This should be protected but VC++ requires it to be public. 
        enum SceneMode { VIEW_MODE, EDIT_MODE };

        void setEditable(bool on);
        bool isEditable();

        void requestRedraw();
        
      protected:
        inline SceneMode sceneMode() { return sceneMode_; }

        virtual ~OSGSceneObject();

        virtual void onAttachedToScene();
        virtual void onDetachedFromScene();

        virtual bool onKeyPressEvent(const OSGSceneViewEvent& event);
        virtual bool onKeyReleaseEvent(const OSGSceneViewEvent& event);
        virtual bool onButtonPressEvent(const OSGSceneViewEvent& event);
        virtual bool onButtonReleaseEvent(const OSGSceneViewEvent& event);
        virtual bool onDoubleClickEvent(const OSGSceneViewEvent& event);
        virtual bool onPointerMoveEvent(const OSGSceneViewEvent& event);
        virtual void onPointerLeaveEvent(const OSGSceneViewEvent& event);
        virtual bool onScrollEvent(const OSGSceneViewEvent& event);
        virtual void onContextMenuRequest(const OSGSceneViewEvent& event, MenuManager& menuManager);
        virtual void onSceneModeChanged();
        virtual bool onUndoRequest();
        virtual bool onRedoRequest();

      private:
        bool isActive_; // in the scene graph ?
        bool isEditable_;
        SceneMode sceneMode_;

        friend class OSGSceneViewImpl;
    };

    typedef osg::ref_ptr<OSGSceneObject> OSGSceneObjectPtr;
}

#endif
