/**
   @author Shizuko Hattori
*/

#include <boost/bind.hpp>
#include "ScenePieces.h"

using namespace std;
using namespace cnoid;

SgCrossMarker::SgCrossMarker(const Vector4& color, float size, float lineWidth)
    : color(color),
      size(size),
      lineWidth(lineWidth)
{
    setRenderingFunction(boost::bind(&SgCrossMarker::renderCrossMarker, this));
}


void SgCrossMarker::setPosition(const Vector3& p)
{
    position = p;
}


void SgCrossMarker::renderCrossMarker()
{
    glPushAttrib(GL_LIGHTING_BIT | GL_CURRENT_BIT);
    glDisable(GL_LIGHTING);

    glColor3f(color[0], color[1], color[2]);
    glLineWidth(lineWidth);
    glBegin(GL_LINES);

    SgFloat x = position.x();
    SgFloat y = position.y();
    SgFloat z = position.z();
    
    glVertex3d(x-size, y, z);
    glVertex3d(x+size, y, z);

    glVertex3d(x, y-size, z);
    glVertex3d(x, y+size, z);

    glVertex3d(x, y, z-size);
    glVertex3d(x, y, z+size);

    glEnd();
    glPopAttrib();

}

SgSphereMarker::SgSphereMarker(float radius, const Vector4& color)
{
    cross = new SgCrossMarker( color, radius*2 );
    addChild( cross );
    SgMeshPtr mesh = new SgMesh;
    mesh->setSphere(radius);
    SgMaterialPtr material = new SgMaterial;
    SgVector3 c(color[0], color[1], color[2]);
    material->setDiffuseColor(c);
    material->setEmissiveColor(c);
    material->setTransparency(1.0 - color[3]);
    SgShapePtr shape = new SgShape;
    shape->setMesh(mesh);
    shape->setMaterial(material);
    addChild(shape);
}


void SgSphereMarker::setCross(const Vector4& color, float size, float lineWidth)
{
    if(cross)
        removeChild(cross);
    cross = new SgCrossMarker( color, size, lineWidth);
    addChild(cross);
}


void SgSphereMarker::setPosition(const Vector3& p)
{
    cross->setPosition(Vector3(0.0, 0.0, 0.0));
    Affine3d T;
    T.linear().setIdentity();
    T.translation() = p;
    setTransform(T);
}


SgBBMarker::SgBBMarker(const SgBoundingBox& bbox, const Vector4& color, float width)
{
    create(bbox, color, width);
}


SgBBMarker::SgBBMarker(const SgBoundingBox& bbox, const Vector4& color)
{
    double h = bbox.max().x() - bbox.min().x();
    double w = bbox.max().y() - bbox.min().y();
    double d = bbox.max().z() - bbox.min().z();
    double width = (h + w + d) / 3.0 / 10.0;
    create(bbox, color, width);
}


void SgBBMarker::create(const SgBoundingBox& bbox, const Vector4& color, float width)
{
    SgMeshPtr mesh = new SgMesh;
    mesh->setBox(SgVector3(width, width, width));
    SgMaterialPtr material = new SgMaterial;
    SgVector3 c(color[0], color[1], color[2]);
    material->setDiffuseColor(c);
    material->setEmissiveColor(c);
    material->setTransparency(1.0 - color[3]);
    SgShapePtr shape = new SgShape;
    shape->setMesh(mesh);
    shape->setMaterial(material);

    const double& x0 = bbox.min().x();
    const double& x1 = bbox.max().x();
    const double& y0 = bbox.min().y();
    const double& y1 = bbox.max().y();
    const double& z0 = bbox.min().z();
    const double& z1 = bbox.max().z();
    
    addMarker(x0, y0, z0, shape);
    addMarker(x0, y0, z1, shape);
    addMarker(x0, y1, z0, shape);
    addMarker(x0, y1, z1, shape);

    addMarker(x1, y0, z0, shape);
    addMarker(x1, y0, z1, shape);
    addMarker(x1, y1, z0, shape);
    addMarker(x1, y1, z1, shape);
}


void SgBBMarker::addMarker(double x, double y, double z, SgShapePtr shape)
{
    Affine3d T;
    T.linear().setIdentity();
    T.translation() << x, y, z;
    SgTransformPtr transform = new SgTransform;
    transform->setTransform(T);

    transform->addChild(shape);
    addChild(transform);
}


SgAttitudeDragger::SgAttitudeDragger()
{
    currentParent = 0;

    static const int numDivisions = 30;
    static const float beltWidthRatio = 0.1f;
    static const int axes[3][3] = { { 0, 1, 2 }, { 1, 0, 2 }, { 2, 1, 0 } };
    static const char* axisNames[3] = { "x", "y", "z" };
    static const double PI = 3.14159265358979323846;

    // make dragger belts
    for(int i=0; i < 3; ++i){
        
        SgMaterialPtr material = new SgMaterial;
        SgVector3 color(0.2f, 0.2f, 0.2f);
        color[i] = 1.0f;
        material->setDiffuseColor(SgVector3(0.0, 0.0, 0.0));
        material->setEmissiveColor(color);
        material->setAmbientIntensity(0.0);
        material->setTransparency(0.6f);

        const int x = axes[i][0];
        const int y = axes[i][1];
        const int z = axes[i][2];

        SgMeshPtr mesh = new SgMesh;
        SgVertexArrayPtr vertices = new SgVertexArray();
        vertices->reserve(numDivisions*2);
        mesh->setVertices(vertices);

        mesh->triangles().clear();
        mesh->triangles().reserve(numDivisions*2);
        for(int j=0; j < numDivisions; ++j){

            const double theta = j * 2.0 * PI / numDivisions;
            SgVector3 v1, v2;
            v1[x] = beltWidthRatio;
            v2[x] = -beltWidthRatio;
            v1[y] = v2[y] = cos(theta);
            v1[z] = v2[z] = sin(theta);
            vertices->push_back(v1);
            vertices->push_back(v2);

            const int s = j * 2;
            const int t = ((j + 1) % numDivisions) * 2;
            mesh->addTriangle(s, t, t+1);
            mesh->addTriangle(s, t+1, s+1);
        }
        mesh->generateNormals(PI);

        draggers[i] = new SgShape;
        draggers[i]->setMesh(mesh);
        draggers[i]->setMaterial(material);
        draggers[i]->setName(axisNames[i]);

        addChild(draggers[i]);
    }
}


void SgAttitudeDragger::attachTo(SgTransformPtr parent)
{
    if(currentParent){
        currentParent->removeChild(this);
        currentParent = 0;
    }
    if(parent){
        float r = parent->untransformedBoundingBox().boundingSphereRadius();
        T() = Eigen::Scaling(r, r, r);
        parent->addChild(this);
        currentParent = parent.get();
    }
}
