/**
   \file
   \author Shin'ichiro Nakaoka
*/

#include "Body.h"
#include "BodyCustomizerInterface.h"
#include <cnoid/EigenUtil>
#include <cnoid/ColdetModel>
#include <boost/make_shared.hpp>

using namespace std;
using namespace boost;
using namespace cnoid;

namespace {

    const bool PUT_DEBUG_MESSAGE = true;

    bool pluginsInDefaultDirectoriesLoaded = false;

#ifndef uint
    typedef unsigned int uint;
#endif
}


Body::Body()
{
    initialize();
    rootLink_ = createLink();
    centerOfMass_.setZero();
    mass_ = 0.0;
    info_ = new Mapping();
    defaultRootPosition.setZero();
    defaultRootAttitude.setIdentity();
}


void Body::initialize()
{
    rootLink_ = 0;
    customizerHandle_ = 0;
    customizerInterface_ = 0;
    bodyHandleEntity.body = this;
    bodyHandle = &bodyHandleEntity;
}


Body::Body(const Body& org)
{
    copy(org);
}


void Body::copy(const Body& org)
{
    initialize();

    name_ = org.name_;
    modelName_ = org.modelName_;
    info_ = org.info_;
    centerOfMass_ = org.centerOfMass_;
    mass_ = org.mass_;

    setRootLink(cloneLinkTree(org.rootLink()));

    // deep copy of the devices
    const DeviceList<>& orgDevices = org.devices();
    for(size_t i=0; i < orgDevices.size(); ++i){
        Device* orgDevice = orgDevices.get(i);
        Device* device = orgDevice->clone();
        device->setLink(link(orgDevice->link()->index()));
        addDevice(device);
    }

    // deep copy of the extraJoints
    for(size_t i=0; i < org.extraJoints.size(); ++i){
        const ExtraJoint& orgExtraJoint = org.extraJoints[i];
        ExtraJoint extraJoint(orgExtraJoint);
        for(int j=0; j < 2; ++j){
            extraJoint.link[j] = link(orgExtraJoint.link[j]->index());
        }
        extraJoints.push_back(extraJoint);
    }

    if(org.customizerInterface_){
        installCustomizer(org.customizerInterface_);
    }
}


Link* Body::cloneLinkTree(const Link* orgLink)
{
    Link* link = createLink(orgLink);
    for(Link* orgChild = orgLink->child(); orgChild; orgChild = orgChild->sibling()){
        link->appendChild(cloneLinkTree(orgChild));
    }
    return link;
}


BodyPtr Body::clone() const
{
    return make_shared<Body>(*this);
}


Link* Body::createLink(const Link* org) const
{
    return org ? new Link(*org) : new Link();
}


Body::~Body()
{
    if(customizerHandle_){
        customizerInterface_->destroy(customizerHandle_);
    }
    delete rootLink_;
}


void Body::setRootLink(Link* link)
{
    if(rootLink_){
        delete rootLink_;
    }
    rootLink_ = link;

    if(rootLink_){
        updateLinkTree();
    }
}


Link* Body::createEmptyJoint(int jointId)
{
    Link* empty = createLink();
    empty->setJointId(jointId);
    return empty;
}


void Body::updateLinkTree()
{
    isStaticModel_ = true;
    
    nameToLinkMap.clear();
    linkTraverse_.find(rootLink());

    const int n = linkTraverse_.numLinks();
    jointIdToLinkArray.clear();
    jointIdToLinkArray.resize(n, 0);
    int maxJointID = -1;

    mass_ = 0.0;
    
    for(int i=0; i < n; ++i){
        Link* link = linkTraverse_[i];
        link->setIndex(i);
        nameToLinkMap[link->name()] = link;

        const int id = link->jointId();
        if(id >= 0){
            int size = jointIdToLinkArray.size();
            if(id >= size){
                jointIdToLinkArray.resize(id + 1, 0);
            }
            jointIdToLinkArray[id] = link;
            if(id > maxJointID){
                maxJointID = id;
            }
        }

        if(link->jointType() != Link::FIXED_JOINT){
            isStaticModel_ = false;
        }
        mass_ += link->mass();
    }

    jointIdToLinkArray.resize(maxJointID + 1);

    for(size_t i=0; i < jointIdToLinkArray.size(); ++i){
        if(!jointIdToLinkArray[i]){
            jointIdToLinkArray[i] = createEmptyJoint(i);
        }
    }
}


void Body::initializeState()
{
    rootLink_->T() = rootLink_->Tb();

    rootLink_->v().setZero();
    rootLink_->w().setZero();
    rootLink_->dv().setZero();
    rootLink_->dw().setZero();
    
    const int n = linkTraverse_.numLinks();
    for(int i=0; i < n; ++i){
        Link* link = linkTraverse_[i];
        link->u() = 0.0;
        link->q() = 0.0;
        link->dq() = 0.0;
        link->ddq() = 0.0;
    }
 
    calcForwardKinematics(true, true);
    clearExternalForces();
    initializeDeviceStates();
}


#ifdef CNOID_BACKWARD_COMPATIBILITY

double Body::calcTotalMass()
{
    mass_ = 0.0;

    int n = linkTraverse_.numLinks();
    for(int i=0; i < n; ++i){
        mass_ += linkTraverse_[i]->m();
    }

    return mass_;
}

void Body::setDefaultRootPosition(const Vector3& p, const Matrix3& R)
{
    rootLink_->setOffsetTranslation(p);
    rootLink_->setOffsetRotation(R);
}


void Body::getDefaultRootPosition(Vector3& out_p, Matrix3& out_R)
{
    out_p = rootLink_->offsetTranslation();
    out_R = rootLink_->offsetRotation();
}

#endif



void Body::resetInfo(MappingPtr info)
{
    info_ = info;
    //doResetInfo(*info);
}


const Vector3& Body::calcCenterOfMass()
{
    mass_ = 0.0;
    Vector3 mc = Vector3::Zero();
    int n = linkTraverse_.numLinks();
    
    for(int i=0; i < n; i++){
        Link* link = linkTraverse_[i];
        link->wc().noalias() = link->R() * link->c() + link->p();
        mc.noalias() += link->m() * link->wc();
        mass_ += link->m();
    }

    centerOfMass_ = mc / mass_;

    return centerOfMass_;
}


/**
   assuming Link::v,w is already computed by calcForwardKinematics(true);
   assuming Link::wc is already computed by calcCenterOfMass();
*/
void Body::calcTotalMomentum(Vector3& out_P, Vector3& out_L)
{
    out_P.setZero();
    out_L.setZero();

    Vector3 dwc;    // Center of mass speed in world frame
    Vector3 P;	    // Linear momentum of the link
    Vector3 L;	    // Angular momentum with respect to the world frame origin 
    Vector3 Llocal; // Angular momentum with respect to the center of mass of the link

    int n = linkTraverse_.numLinks();
    
    for(int i=0; i < n; i++){
        Link* link = linkTraverse_[i];
        dwc = link->v() + link->w().cross(link->R() * link->c());
        P   = link->m() * dwc;

        //L   = cross(link->wc, P) + link->R * link->I * trans(link->R) * link->w; 
        Llocal.noalias() = link->I() * link->R().transpose() * link->w();
        L     .noalias() = link->wc().cross(P) + link->R() * Llocal; 

        out_P += P;
        out_L += L;
    }
}


void Body::addDevice(const DevicePtr& device)
{
    devices_.push_back(device);
    if(!device->name().empty()){
        deviceNameMap[device->name()] = device.get();
    }
}


void Body::initializeDeviceStates()
{
    for(size_t i=0; i < devices_.size(); ++i){
        devices_[i]->initializeState();
    }
}


bool Body::hasVirtualJointForces() const
{
    return (customizerInterface_ && customizerInterface_->setVirtualJointForces);
}


void Body::setVirtualJointForces()
{
    if(customizerInterface_ && customizerInterface_->setVirtualJointForces){
        customizerInterface_->setVirtualJointForces(customizerHandle_);
    }
}


void Body::clearExternalForces()
{
    int n = linkTraverse_.numLinks();
    for(int i=0; i < n; ++i){
        Link* link = linkTraverse_[i];
        link->F_ext().setZero();
    }
}


void Body::updateLinkColdetModelPositions()
{
    const int n = linkTraverse_.numLinks();
    for(int i=0; i < n; ++i){
        Link* link = linkTraverse_[i];
        if(link->coldetModel()){
            link->coldetModel()->setPosition(link->T());
        }
    }
}


/**
   The function installs the pre-loaded customizer corresponding to the model name.
*/
bool Body::installCustomizer()
{
    if(!pluginsInDefaultDirectoriesLoaded){
        loadBodyCustomizers(bodyInterface());
        pluginsInDefaultDirectoriesLoaded = true;
    }
		
    BodyCustomizerInterface* interface = findBodyCustomizer(modelName_);

    return interface ? installCustomizer(interface) : false;
}


bool Body::installCustomizer(BodyCustomizerInterface* customizerInterface)
{
    if(customizerInterface_){
        if(customizerHandle_){
            customizerInterface_->destroy(customizerHandle_);
            customizerHandle_ = 0;
        }
        customizerInterface_ = 0;
    }
	
    if(customizerInterface){
        customizerHandle_ = customizerInterface->create(bodyHandle, modelName_.c_str());
        if(customizerHandle_){
            customizerInterface_ = customizerInterface;
        }
    }

    return (customizerHandle_ != 0);
}


static inline Link* extractLink(BodyHandle bodyHandle, int linkIndex)
{
    return static_cast<BodyHandleEntity*>(bodyHandle)->body->link(linkIndex);
}


static int getLinkIndexFromName(BodyHandle bodyHandle, const char* linkName)
{
    Body* body = static_cast<BodyHandleEntity*>(bodyHandle)->body;
    Link* link = body->link(linkName);
    return (link ? link->index() : -1);
}


static const char* getLinkName(BodyHandle bodyHandle, int linkIndex)
{
    return extractLink(bodyHandle, linkIndex)->name().c_str();
}


static double* getJointValuePtr(BodyHandle bodyHandle, int linkIndex)
{
    return &(extractLink(bodyHandle,linkIndex)->q());
}


static double* getJointVelocityPtr(BodyHandle bodyHandle, int linkIndex)
{
    return &(extractLink(bodyHandle, linkIndex)->dq());
}


static double* getJointTorqueForcePtr(BodyHandle bodyHandle, int linkIndex)
{
    return &(extractLink(bodyHandle, linkIndex)->u());
}


BodyInterface* Body::bodyInterface()
{
    static BodyInterface interface = {
        BODY_INTERFACE_VERSION,
        getLinkIndexFromName,
        getLinkName,
        getJointValuePtr,
        getJointVelocityPtr,
        getJointTorqueForcePtr,
    };

    return &interface;
}
