/**
   \file
   \author Shin'ichiro Nakaoka
*/

#include "BodyLoader.h"
#include "Sensor.h"
#include "Camera.h"
#include "Light.h"
#include "ModelNodeSet.h"
#include <cnoid/Exception>
#include <cnoid/EigenTypes>
#include <cnoid/YAMLReader>
#include <cnoid/TriangleMeshShaper>
#include <cnoid/VRMLToSGConverter>
#include <cnoid/ColdetModel>
#include <cnoid/FileUtil>
#include <Eigen/StdVector>
#include <boost/bind.hpp>
#include <boost/function.hpp>
#include <boost/format.hpp>
#include <boost/dynamic_bitset.hpp>
#include <boost/make_shared.hpp>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

namespace cnoid {

    class BodyLoaderImpl
    {
    public:
        BodyLoaderImpl();
        ~BodyLoaderImpl();

        void setDivisionNumber(int n);

        bool load(Body& body, const std::string& filename);
        BodyPtr load(const std::string& filename);        
        
#ifdef CNOID_BACKWARD_COMPATIBILITY
        BodyPtr loadModelFile(
            const std::string& filename, bool doTriangulation, bool doNormalGeneration, bool createColdetModel);
#endif
        ModelNodeSetPtr modelNodeSet;
        VRMLProtoInstancePtr rootJointNode;
        TriangleMeshShaper triangleMeshShaper;
        bool doTriangulation;
        bool doNormalGeneration;
        bool doOverwriteNormals;
        bool doColdetModelCreation;
        string errorMessage;
        MappingPtr info;
        dynamic_bitset<> validJointIdSet;
        int numValidJointIds;
        VRMLToSGConverter sgConverter;

        boost::signal<void(const std::string& message)> sigMessage;

        typedef boost::function<DevicePtr(VRMLProtoInstance* node)> DeviceFactory;
        
        typedef map<string, DeviceFactory> DeviceFactoryMap;
        static DeviceFactoryMap deviceFactories;

        struct MeshInfo
        {
            EIGEN_MAKE_ALIGNED_OPERATOR_NEW

            MeshInfo(VRMLIndexedFaceSet* triangleMesh, const Affine3& TX) :
                triangleMesh(triangleMesh), TX(TX) { }
            VRMLIndexedFaceSet* triangleMesh;
            Affine3 TX;
        };
        vector<MeshInfo, Eigen::aligned_allocator<MeshInfo> > meshes;
        int numTotalVertices;
        int numTotalTriangles;

        void constructBodyFromModelNodeSet(Body& body);
        Link* createLink(Body& body, JointNodeSetPtr jointNodeSet, const Matrix3& parentRs);
        void setJointProperties(Link* link, VRMLProtoInstancePtr jointNode, const Matrix3& parentRs);
        void setSegmentProperties(Link* link, vector<VRMLProtoInstancePtr> segmentNodes, const JointNodeSet::Affine3Array& transforms);

        void createDevices(Body& body, Link* link, std::vector<VRMLProtoInstancePtr>& deviceNodes);
        static void readDeviceCommonParameters(Device& device, VRMLProtoInstance* node);
        static ForceSensorPtr createForceSensor(VRMLProtoInstance* node);
        static RateGyroSensorPtr createRateGyroSensor(VRMLProtoInstance* node);
        static AccelSensorPtr createAccelSensor(VRMLProtoInstance* node);
        static void readCameraCommonParameters(Camera& camera, VRMLProtoInstance* node);
        static CameraPtr createCamera(VRMLProtoInstance* node);
        static void readLightDeviceCommonParameters(Light& light, VRMLProtoInstance* node);
        static SpotLightPtr createSpotLight(VRMLProtoInstance* node);
        
        void setColdetModel(Link* link, vector<VRMLProtoInstancePtr> segmentNodes, const JointNodeSet::Affine3Array& transforms);
        bool collectMeshes(VRMLShape* shapeNode, const Affine3& T);
        void extractShapeNodes(VRMLNode* node, const function<bool(VRMLShape* shapeNode, const Affine3& T)>& callback);
        bool extractShapeNodeTraverse(
            VRMLNode* node, const Affine3& T, const function<bool(VRMLShape* shapeNode, const Affine3& T)>& callback);
        void setExtraJoints(Body& body);
    };
}


BodyLoaderImpl::DeviceFactoryMap BodyLoaderImpl::deviceFactories;


namespace {

    double getLimitValue(VRMLVariantField& field, double defaultValue)
    {
        MFFloat& values = get<MFFloat>(field);
        if(values.empty()){
            return defaultValue;
        }
        return values[0];
    }

    template<class ValueType> ValueType getValue(VRMLProtoInstance* node, const char* fieldName)
    {
        VRMLProtoFieldMap::const_iterator p = node->fields.find(fieldName);
        if(p == node->fields.end()){
            BOOST_THROW_EXCEPTION(
                nonexistent_key_error()
                << error_info_key(fieldName)
                << error_info_message(str(format(_("Node \"%1%\" should have the field \"%2%\"")) % node->proto->protoName % fieldName)));
        }
        return boost::get<ValueType>(p->second);
    }
        
    void readVRMLfield(VRMLVariantField& field, string& out_s)
    {
        switch(field.which()){
        case SFSTRING:
            out_s = get<SFString>(field);
            break;
        case MFSTRING:
        {
            MFString& strings = get<MFString>(field);
            out_s = "";
            for(size_t i=0; i < strings.size(); i++){
                out_s += strings[i] + "\n";
            }
        }
        break;
        default:
            break;
        }
    }


    bool checkAndReadVRMLfield(VRMLProtoInstance* node, const char* key, bool& out_value)
    {
        VRMLVariantField* field = node->findField(key);
        if(field && field->which() == SFBOOL){
            out_value = get<SFBool>(*field);
            return true;
        }
        return false;
    }

    void readVRMLfield(VRMLVariantField& field, int& out_value)
    {
        out_value = get<SFInt32>(field);
    }

    bool checkAndReadVRMLfield(VRMLProtoInstance* node, const char* key, int& out_value)
    {
        VRMLVariantField* field = node->findField(key);
        if(field && field->which() == SFINT32){
            out_value = get<SFInt32>(*field);
            return true;
        }
        return false;
    }


    void readVRMLfield(VRMLVariantField& field, double& out_value)
    {
        out_value = get<SFFloat>(field);
    }


    bool checkAndReadVRMLfield(VRMLProtoInstance* node, const char* key, SFVec3f& out_value)
    {
        VRMLVariantField* field = node->findField(key);
        if(field && field->which() == SFVEC3F){
            out_value = get<SFVec3f>(*field);
            return true;
        }
        return false;
    }
    

    void readVRMLfield(VRMLVariantField& field, Vector3& out_v)
    {
        out_v = get<SFVec3f>(field);
    }


    void readVRMLfield(VRMLVariantField& field, Matrix3& out_R)
    {
        if(field.which() == SFROTATION){
            out_R = get<SFRotation>(field).toRotationMatrix();

        } else if(field.which() == MFFLOAT){
            MFFloat& mf = get<MFFloat>(field);
            if(mf.size() >= 9){
                out_R <<
                    mf[0], mf[1], mf[2],
                    mf[3], mf[4], mf[5],
                    mf[6], mf[7], mf[8];
            }
        }
    }
}


BodyLoader::BodyLoader()
{
    impl = new BodyLoaderImpl();
}


BodyLoaderImpl::BodyLoaderImpl()
{
    doTriangulation = true;
    doNormalGeneration = true;
    doOverwriteNormals = false;
    doColdetModelCreation = true;
    
    if(deviceFactories.empty()){
        deviceFactories["ForceSensor"]        = &BodyLoaderImpl::createForceSensor;
        deviceFactories["Gyro"]               = &BodyLoaderImpl::createRateGyroSensor;
        deviceFactories["AccelerationSensor"] = &BodyLoaderImpl::createAccelSensor;
        //sensorTypeMap["PressureSensor"]     = Sensor::PRESSURE;
        //sensorTypeMap["PhotoInterrupter"]   = Sensor::PHOTO_INTERRUPTER;
        //sensorTypeMap["TorqueSensor"]       = Sensor::TORQUE;
        //sensorTypeMap["RangeSensor"]        = Sensor::RANGE;
        deviceFactories["VisionSensor"]       = &BodyLoaderImpl::createCamera;
        deviceFactories["SpotLightDevice"]    = &BodyLoaderImpl::createSpotLight;
    }

    triangleMeshShaper.sigMessage.connect(sigMessage);
}


BodyLoader::~BodyLoader()
{
    delete impl;
}


BodyLoaderImpl::~BodyLoaderImpl()
{

}


ModelNodeSetPtr BodyLoader::modelNodeSet()
{
    return impl->modelNodeSet;
}


const std::string& BodyLoader::errorMessage()
{
    return impl->errorMessage;
}


SignalProxy< boost::signal<void(const std::string& message)> > BodyLoader::sigMessage()
{
    return impl->sigMessage;
}


/**
   \todo fully implement this mode
*/
void BodyLoader::enableShapeLoading(bool on)
{
    if(!on){
        enableTriangulation(false);
        enableNormalGeneration(false);
        enableColdetModelCreation(false);
    }
}
    

/**
  The function for setting the division number of primitive geometries such as
   cone, cylinder and sphere.
*/
void BodyLoader::setDivisionNumber(int n)
{
    impl->setDivisionNumber(n);
}


void BodyLoaderImpl::setDivisionNumber(int n)
{
    triangleMeshShaper.setDivisionNumber(std::max(4, n));
}


void BodyLoader::enableTriangulation(bool on)
{
    impl->doTriangulation = on;
}


void BodyLoader::enableNormalGeneration(bool on, bool doOverwrite)
{
    impl->doNormalGeneration = on;
    impl->doOverwriteNormals = doOverwrite;
}


void BodyLoader::enableColdetModelCreation(bool on)
{
    impl->doColdetModelCreation = on;
}


bool BodyLoader::load(Body& body, const std::string& filename)
{
    return impl->load(body, filename);
}


BodyPtr BodyLoader::load(const std::string& filename)
{
    BodyPtr body = make_shared<Body>();
    if(impl->load(*body, filename)){
        return body;
    }
    return BodyPtr();
}


#ifdef CNOID_BACKWARD_COMPATIBILITY
BodyPtr BodyLoader::loadModelFile
(const std::string& filename, bool doTriangulation, bool doNormalGeneration, bool createColdetModel)
{
    enableTriangulation(doTriangulation);
    enableNormalGeneration(doNormalGeneration);
    enableColdetModelCreation(createColdetModel);
    
    BodyPtr body = make_shared<Body>();
    if(!impl->load(*body, filename)){
        body.reset();
    }
    return body;
}
#endif

bool BodyLoaderImpl::load(Body& body, const std::string& filename)
{
    bool result = false;
    
    info = 0;
    validJointIdSet.clear();
    numValidJointIds = 0;
    errorMessage.clear();
    
    filesystem::path orgpath(filename);
    string ext = filesystem::extension(orgpath);
    
    string vrmlFile;

    try {
        if(ext == ".wrl"){
            vrmlFile = filename;

        } else if(ext == ".yaml"){
            YAMLReader parser;
            info = parser.loadDocument(filename)->toMapping();
            filesystem::path vrmlFilePath(info->get("modelFile").toString());
            if(vrmlFilePath.has_root_path()){
                vrmlFile = getNativePathString(vrmlFilePath);
            } else {
                vrmlFile = getNativePathString(orgpath.parent_path() / vrmlFilePath);
            }
        }
        
        if(!vrmlFile.empty()){
            if(info){
                int divisionNumber;
                if(info->read("divisionNumberOfPrimitiveGeometries", divisionNumber)){
                    setDivisionNumber(divisionNumber);
                }
            }

            modelNodeSet = make_shared<ModelNodeSet>();
            
            if(modelNodeSet->loadModelFile(vrmlFile)){
                if(doTriangulation || doColdetModelCreation){
                    if(doOverwriteNormals){
                        triangleMeshShaper.enableNormalGeneration(doNormalGeneration, true);
                        triangleMeshShaper.setMinimumCreaseAngle(1.2);
                    } else {
                        triangleMeshShaper.enableNormalGeneration(doNormalGeneration);
                    }
                    triangleMeshShaper.apply(modelNodeSet->humanoidNode());
                }

                constructBodyFromModelNodeSet(body);
                result = true;
            }
        }
        
    } catch(const ValueNode::Exception& ex){
        errorMessage = ex.message();
    } catch(ModelNodeSet::Exception& ex){
        errorMessage = ex.what();
    } catch(const nonexistent_key_error& error){
        if(const std::string* message = get_error_info<error_info_message>(error)){
            errorMessage = *message;
        }
    } catch(const std::exception& ex){
        errorMessage = ex.what();
    }
    
    return result;
}


void BodyLoaderImpl::constructBodyFromModelNodeSet(Body& body)
{
    JointNodeSetPtr rootJointNodeSet = modelNodeSet->rootJointNodeSet();

    if(rootJointNodeSet){

        body.setModelName(modelNodeSet->humanoidNode()->defName);

        rootJointNode = rootJointNodeSet->jointNode;
    
        Matrix3 Rs = Matrix3::Identity();
        Link* rootLink = createLink(body, rootJointNodeSet, Rs);

        // Warn empty joint ids
        if(numValidJointIds < validJointIdSet.size()){
            for(size_t i=0; i < validJointIdSet.size(); ++i){
                if(!validJointIdSet[i]){
                    sigMessage(str(format("Warning: Joint ID %1% is not specified.\n") % i));
                }
            }
        }
        
        VRMLProtoFieldMap& f = rootJointNodeSet->jointNode->fields;

        Vector3 defaultRootPos;
        readVRMLfield(f["translation"], defaultRootPos);
        Matrix3 defaultRootR;
        readVRMLfield(f["rotation"], defaultRootR);

        rootLink->setOffsetTranslation(defaultRootPos);
        rootLink->setOffsetRotation(defaultRootR);

        body.setRootLink(rootLink);

        body.installCustomizer();

        setExtraJoints(body);

        if(info){
            body.resetInfo(info);
        }
    }
}


Link* BodyLoaderImpl::createLink(Body& body, JointNodeSetPtr jointNodeSet, const Matrix3& parentRs)
{
    Link* link = body.createLink();

    setJointProperties(link, jointNodeSet->jointNode, parentRs);

    vector<VRMLProtoInstancePtr> segmentNodes = jointNodeSet->segmentNodes;

    const JointNodeSet::Affine3Array& segmentTransforms = jointNodeSet->segmentTransforms;

    setSegmentProperties(link, segmentNodes, segmentTransforms);
    
    if(doColdetModelCreation && !segmentNodes.empty()){
        setColdetModel(link, segmentNodes, segmentTransforms);
    }
    
    createDevices(body, link, jointNodeSet->deviceNodes);

    const int n = jointNodeSet->childJointNodeSets.size();
    for(int i = 0; i < n; ++i){
        JointNodeSetPtr& childJointNodeSet = jointNodeSet->childJointNodeSets[i];
        link->appendChild(createLink(body, childJointNodeSet, link->Rs()));
    }

    return link;
}


void BodyLoaderImpl::setJointProperties(Link* link, VRMLProtoInstancePtr jointNode, const Matrix3& parentRs)
{
    link->setName(jointNode->defName);

    VRMLProtoFieldMap& jf = jointNode->fields;

    link->setJointId(get<SFInt32>(jf["jointId"]));

    if(link->jointId() >= 0){
        if(link->jointId() >= validJointIdSet.size()){
            validJointIdSet.resize(link->jointId() + 1);
        }
        if(!validJointIdSet[link->jointId()]){
            ++numValidJointIds;
            validJointIdSet.set(link->jointId());
        } else {
            sigMessage(str(format("Warning: Joint ID %1% is duplicated.\n") % link->jointId()));
        }
    }

    Vector3 b;
    readVRMLfield(jf["translation"], b);
    link->setOffsetTranslation(parentRs * b);

    if(jointNode != rootJointNode){
        Matrix3 R;
        readVRMLfield(jf["rotation"], R);
        link->setAccumlatedSegmentRotation(parentRs * R);
    } else {
        link->setAccumlatedSegmentRotation(parentRs);
    }

    string jointType;
    readVRMLfield(jf["jointType"], jointType);
    
    if(jointType == "fixed" ){
        link->setJointType(Link::FIXED_JOINT);
    } else if(jointType == "free" ){
        link->setJointType(Link::FREE_JOINT);
    } else if(jointType == "rotate" ){
        link->setJointType(Link::ROTATIONAL_JOINT);
    } else if(jointType == "slide" ){
        link->setJointType(Link::SLIDE_JOINT);
    } else if(jointType == "crawler"){
        link->setJointType(Link::CRAWLER_JOINT);
    } else {
        link->setJointType(Link::FIXED_JOINT);
    }

    if(link->jointType() == Link::FREE_JOINT || link->jointType() == Link::FIXED_JOINT){
        link->setJointAxis(Vector3::Zero());

    } else {
        Vector3 jointAxis;
        VRMLVariantField& jointAxisField = jf["jointAxis"];
        switch(jointAxisField.which()){
        case SFSTRING:
        {
            SFString& axisLabel = get<SFString>(jointAxisField);
            if(axisLabel == "X"){
                jointAxis = Vector3::UnitX();
            } else if(axisLabel == "Y"){
                jointAxis = Vector3::UnitY();
            } else if(axisLabel == "Z"){
                jointAxis = Vector3::UnitZ();
            }
        }
        break;
        case SFVEC3F:
            readVRMLfield(jointAxisField, jointAxis);
            break;
        default:
            jointAxis = Vector3::UnitZ();
            break;
        }
        link->setJointAxis(link->Rs() * jointAxis);
    }

    //readVRMLfield(jf["jointValue"], link->defaultJointValue);

    double Ir, gearRatio, torqueConst, encoderPulse, rotorResistor;
    readVRMLfield(jf["rotorInertia"], Ir);
    readVRMLfield(jf["gearRatio"], gearRatio);
    readVRMLfield(jf["torqueConst"], torqueConst);
    readVRMLfield(jf["encoderPulse"], encoderPulse);
    readVRMLfield(jf["rotorResistor"], rotorResistor);

    VRMLVariantField* field = jointNode->findField("equivalentInertia");
    if(field){
        link->setEquivalentRotorInertia(get<SFFloat>(*field));
    } else {
        link->setEquivalentRotorInertia(gearRatio * gearRatio * Ir);
    }    

    double maxlimit = numeric_limits<double>::max();

    link->setJointRange(
        getLimitValue(jf["llimit"],  -maxlimit),
        getLimitValue(jf["ulimit"],  +maxlimit));

    link->setJointVelocityRange(
        getLimitValue(jf["lvlimit"], -maxlimit),
        getLimitValue(jf["uvlimit"], +maxlimit));
}    


void BodyLoaderImpl::setSegmentProperties
(Link* link, vector<VRMLProtoInstancePtr> segmentNodes, const JointNodeSet::Affine3Array& transforms)
{
    const int numSegments = segmentNodes.size();
    double totalMass = 0.0;

    //  Mass = Sigma mass                 //
    //  C = (Sigma mass * T * c) / Mass   //
    //  I = Sigma(R * I * Rt + G)       //
    //  R = Rotation matrix part of T     //
    //  G = y*y+z*z, -x*y, -x*z, -y*x, z*z+x*x, -y*z, -z*x, -z*y, x*x+y*y    //
    //  (x, y, z ) = T * c - C        //
    std::vector<Vector3> centerOfMassArray;
    std::vector<double> massArray;
    Vector3 linkc = Vector3::Zero();
    Matrix3 linkI = Matrix3::Zero();

    if(numSegments > 0){
        for(int i = 0; i < numSegments; ++i){
            const Affine3& T = transforms[i];
            VRMLProtoFieldMap& sf = segmentNodes[i]->fields;
            Vector3 c;
            readVRMLfield(sf["centerOfMass"], c);
            double m;
            readVRMLfield(sf["mass"], m);
            Matrix3 I;
            readVRMLfield(sf["momentsOfInertia"], I);
            
            const Vector3 c1 = T.linear() * c + T.translation();
            centerOfMassArray.push_back(c1);
            massArray.push_back(m);
            linkc = c1 * m + linkc * totalMass;
            totalMass += m;
            linkc /= totalMass;
            linkI.noalias() += T.linear() * I * T.linear().transpose();
        }

        for(int i = 0; i < numSegments; ++i){
            const Vector3& c = centerOfMassArray[i];
            double x = c(0) - linkc(0);
            double y = c(1) - linkc(1);
            double z = c(2) - linkc(2);
            double m = massArray[i];
            
            linkI(0,0) += m * (y*y + z*z);
            linkI(0,1) += -m * x * y;
            linkI(0,2) += -m * x * z;
            linkI(1,0) += -m * y * x;
            linkI(1,1) += m * (z*z + x*x);
            linkI(1,2) += -m * y * z;
            linkI(2,0) += -m * z * x;
            linkI(2,1) += -m * z * y;
            linkI(2,2) += m * (x*x + y*y);
        }

        SgInvariantGroupPtr group = new SgInvariantGroup();
        for(int i=0; i < numSegments; ++i){
            VRMLProtoInstancePtr& segmentNode = segmentNodes[i];
            SgNodePtr node = sgConverter.convert(segmentNode);
            if(node){
                const Affine3& T = transforms[i];
                if(T.matrix() == Affine3::Identity().matrix()){
                    group->addChild(node);
                } else {
                    SgTransform* transform = new SgTransform(T.cast<SgFloat>());
                    transform->addChild(node);
                    group->addChild(transform);
                }
            }
        }
        //if(!group->empty()){
        link->setShape(group);
        //}
    }

    link->setMass(totalMass);
    link->setCenterOfMass(link->Rs() * linkc);
    link->setInertia(link->Rs() * linkI * link->Rs().transpose());
}
    

void BodyLoaderImpl::createDevices(Body& body, Link* link, std::vector<VRMLProtoInstancePtr>& deviceNodes)
{
    const int numDevices = deviceNodes.size();

    for(int i=0; i < numDevices; ++i){
        VRMLProtoInstance* deviceNode = deviceNodes[i].get();

        const string& typeName = deviceNode->proto->protoName;
        DeviceFactoryMap::iterator p = deviceFactories.find(typeName);

        if(p == deviceFactories.end()){
            sigMessage(str(format("Sensor type %1% is not supported.\n") % typeName));

        } else {
            DeviceFactory& factory = p->second;
            DevicePtr device = factory(deviceNode);
            if(device){
                device->setLink(link);
                device->setLocalTranslation(link->Rs() * device->localTranslation());
                device->setLocalRotation(link->Rs() * device->localRotation());
                body.addDevice(device);
            }
        }
    }
}


void BodyLoaderImpl::readDeviceCommonParameters(Device& device, VRMLProtoInstance* node)
{
    device.setName(node->defName);
            
    int id = -1;
    if(!checkAndReadVRMLfield(node, "deviceId", id)){
        checkAndReadVRMLfield(node, "sensorId", id);
    }
    device.setId(id);

    device.setLocalTranslation(getValue<SFVec3f>(node, "translation"));

    Matrix3 R;
    readVRMLfield(node->fields["rotation"], R);
    device.setLocalRotation(R);
}



ForceSensorPtr BodyLoaderImpl::createForceSensor(VRMLProtoInstance* node)
{
    ForceSensorPtr sensor = new ForceSensor();
    readDeviceCommonParameters(*sensor, node);

    SFVec3f f_max, t_max;
    if(checkAndReadVRMLfield(node, "maxForce", f_max)){
        sensor->F_max().head<3>() = f_max;
    }
    if(checkAndReadVRMLfield(node, "maxTorque", t_max)){
        sensor->F_max().tail<3>() = t_max;
    }
    return sensor;
}


RateGyroSensorPtr BodyLoaderImpl::createRateGyroSensor(VRMLProtoInstance* node)
{
    RateGyroSensorPtr sensor = new RateGyroSensor();
    readDeviceCommonParameters(*sensor, node);

    SFVec3f w_max;
    if(checkAndReadVRMLfield(node, "maxAngularVelocity", w_max)){
        sensor->w_max() = w_max;
    }
    return sensor;
}


AccelSensorPtr BodyLoaderImpl::createAccelSensor(VRMLProtoInstance* node)
{
    AccelSensorPtr sensor = new AccelSensor();
    readDeviceCommonParameters(*sensor, node);

    SFVec3f dv_max;
    if(checkAndReadVRMLfield(node, "maxAngularVelocity", dv_max)){
        sensor->dv_max() = dv_max;
    }
    return sensor;
}


void BodyLoaderImpl::readCameraCommonParameters(Camera& camera, VRMLProtoInstance* node)
{
    readDeviceCommonParameters(camera, node);

    bool on = true;
    if(checkAndReadVRMLfield(node, "on", on)){
        camera.on(on);
    }
    camera.setImageWidth(getValue<SFInt32>(node, "width"));
    camera.setImageHeight(getValue<SFInt32>(node, "height"));
    camera.setFieldOfView(getValue<SFFloat>(node, "fieldOfView"));
}


CameraPtr BodyLoaderImpl::createCamera(VRMLProtoInstance* node)
{
    CameraPtr camera = new Camera();
    readCameraCommonParameters(*camera, node);
    return camera;
}


void BodyLoaderImpl::readLightDeviceCommonParameters(Light& light, VRMLProtoInstance* node)
{
    readDeviceCommonParameters(light, node);
    
    light.on(getValue<SFBool>(node, "on"));
    light.setColor(getValue<SFColor>(node, "color").cast<Vector3::Scalar>());
    light.setIntensity(getValue<SFFloat>(node, "intensity"));
}


SpotLightPtr BodyLoaderImpl::createSpotLight(VRMLProtoInstance* node)
{
    SpotLightPtr light = new SpotLight();
    
    readLightDeviceCommonParameters(*light, node);

    light->setDirection(getValue<SFVec3f>(node, "direction"));
    light->setBeamWidth(getValue<SFFloat>(node, "beamWidth"));
    light->setCutOffAngle(getValue<SFFloat>(node, "cutOffAngle"));

    light->setAttenuation(getValue<SFVec3f>(node, "attenuation"));

    return light;
}


void BodyLoaderImpl::setColdetModel
(Link* link, vector<VRMLProtoInstancePtr> segmentNodes, const JointNodeSet::Affine3Array& transforms)
{
    meshes.clear();
    numTotalVertices = 0;
    numTotalTriangles = 0;

    int numSegments = segmentNodes.size();
    for(int i = 0 ; i < numSegments ; ++i){
        extractShapeNodeTraverse(segmentNodes[i].get(), transforms[i], bind(&BodyLoaderImpl::collectMeshes, this, _1, _2));
    }

    ColdetModelPtr coldetModel(make_shared<ColdetModel>());

    coldetModel->setNumVertices(numTotalVertices);
    coldetModel->setNumTriangles(numTotalTriangles);

    int vertexIndex = 0;
    int triangleIndex = 0;
    
    for(size_t i=0; i < meshes.size(); ++i){

        int vertexTop = vertexIndex;

        Affine3 T = Affine3(link->Rs()) * meshes[i].TX;
        
        const MFVec3s& vertices = meshes[i].triangleMesh->coord->point;
        size_t numVertices = vertices.size();
        
        for(size_t j=0; j < numVertices; ++j){
            Vector3 v = T * vertices[j].cast<double>();
            coldetModel->setVertex(vertexIndex++, (float)v[0], (float)v[1], (float)v[2]);
        }

        const MFInt32& indices = meshes[i].triangleMesh->coordIndex;
        const size_t numTriangles = indices.size() / 4;
        for(size_t j=0; j < numTriangles; ++j){
            const int v0 = vertexTop + indices[j*4];
            const int v1 = vertexTop + indices[j*4+1];
            const int v2 = vertexTop + indices[j*4+2];
            coldetModel->setTriangle(triangleIndex++, v0, v1, v2);
        }
    }
    
    coldetModel->setName(link->name());
    coldetModel->build();
    link->setColdetModel(coldetModel);
}


bool BodyLoaderImpl::collectMeshes(VRMLShape* shapeNode, const Affine3& T)
{
    VRMLIndexedFaceSet* triangleMesh = dynamic_node_cast<VRMLIndexedFaceSet>(shapeNode->geometry).get();
    if(triangleMesh){
        meshes.push_back(MeshInfo(triangleMesh, T));
        numTotalVertices += triangleMesh->coord->point.size();
        numTotalTriangles += triangleMesh->coordIndex.size() / 4;
    }

    return true;
}


/**
   @todo move this code into the ModelNodeSet class ?
*/

bool BodyLoaderImpl::extractShapeNodeTraverse
(VRMLNode* node, const Affine3& T, const function<bool(VRMLShape* shapeNode, const Affine3& T)>& callback)
{
    bool doContinue = true;
    
    if(node->isCategoryOf(PROTO_INSTANCE_NODE)){
        VRMLProtoInstance* protoInstance = static_cast<VRMLProtoInstance*>(node);
        if(protoInstance->actualNode){
            doContinue = extractShapeNodeTraverse(protoInstance->actualNode.get(), T, callback);
        }

    } else if(node->isCategoryOf(GROUPING_NODE)) {

        VRMLGroup* groupNode = static_cast<VRMLGroup*>(node);
        
        Affine3 T2;
        VRMLTransform* transformNode = dynamic_cast<VRMLTransform*>(groupNode);
        if(transformNode){
            T2 = T * transformNode->toAffine3d();
        }

        VRMLSwitch* switchNode = dynamic_cast<VRMLSwitch*>(node);
        if(switchNode){
            int whichChoice = switchNode->whichChoice;
            if(whichChoice >= 0 && whichChoice < switchNode->countChildren()){
                extractShapeNodeTraverse(switchNode->getChild(whichChoice), (transformNode ? T2 : T), callback);
            }
        } else {
            for(int i=0; i < groupNode->countChildren(); ++i){
                if(!extractShapeNodeTraverse(groupNode->getChild(i), (transformNode ? T2 : T), callback)){
                    doContinue = false;
                    break;
                }
            }
        }

    } else if(node->isCategoryOf(SHAPE_NODE)) {
        doContinue = callback(static_cast<VRMLShape*>(node), T);
    }

    return doContinue;
}


void BodyLoaderImpl::setExtraJoints(Body& body)
{
    body.extraJoints.clear();
    int n = modelNodeSet->numExtraJointNodes();
    
    for(int i=0; i < n; ++i){

        VRMLProtoFieldMap& f = modelNodeSet->extraJointNode(i)->fields;
        Body::ExtraJoint joint;

        string link1Name, link2Name;
        readVRMLfield(f["link1Name"], link1Name);
        readVRMLfield(f["link2Name"], link2Name);
        joint.link[0] = body.link(link1Name);
        joint.link[1] = body.link(link2Name);

        for(int j=0; j < 2; ++j){
            if(!joint.link[j]){
                throw ModelNodeSet::Exception(
                    str(format("Field \"link%1%Name\" of a ExtraJoint node does not specify a valid link name")
                        % (j+1)));
            }
        }

        SFString& jointType = get<SFString>(f["jointType"]);
        if(jointType == "piston"){
            joint.type = Body::EJ_PISTON;
            joint.axis = get<SFVec3f>(f["jointAxis"]);
        } else if(jointType == "ball"){
            joint.type = Body::EJ_BALL;
        } else {
            throw ModelNodeSet::Exception(str(format("JointType \"%1%\" is not supported.") % jointType));
        }
            
        readVRMLfield(f["link1LocalPos"], joint.point[0]);
        readVRMLfield(f["link2LocalPos"], joint.point[1]);

        body.extraJoints.push_back(joint);
    }
}
