/**
   \file
   \author Shin'ichiro Nakaoka
*/

#ifndef CNOID_BODY_DEVICE_H_INCLUDED
#define CNOID_BODY_DEVICE_H_INCLUDED

#include <cnoid/Referenced>
#include <cnoid/PolymorphicPointerArray>
#include <cnoid/EigenTypes>
#include <cnoid/SignalProxy>
#include <boost/function.hpp>
#include <boost/scoped_ptr.hpp>
#include <string>
#include <typeinfo>
#include "exportdecl.h"

namespace cnoid {

    class Link;


    class CNOID_EXPORT DeviceState : public Referenced
    {
      protected:
        DeviceState() { }
        DeviceState(const DeviceState& org) { }
        
      public:
        virtual ~DeviceState();
        virtual DeviceState* cloneState() = 0;
        virtual bool copyStateFrom(const DeviceState& other) = 0;
    };
    typedef boost::intrusive_ptr<DeviceState> DeviceStatePtr;


    class CNOID_EXPORT Device : public DeviceState
    {
        struct NonState {
            EIGEN_MAKE_ALIGNED_OPERATOR_NEW
            std::string name;
            int id;
            Link* link;
            Isometry3 T_local;
            double cycle;
            /**
               Size of the double-precision floating numbers for representing the state.
               Zero means that the state should be stored as the whole DeviceState.
               This should be initialized in the constructors.
            */
            int stateSize; 
            const Isometry3& const_T_local() const { return T_local; }
        };
        NonState* ns;

        boost::signal<void()> sigStateChanged_;

      protected:
        Device(); 
        Device(const Device& org, bool stateOnly = false);

        void setStateSize(int size) { ns->stateSize = size; }

      public:
        virtual ~Device();
        
        virtual Device* clone() = 0;
        virtual void forEachActualType(boost::function<bool(const std::type_info& type)> func);
        virtual void initializeState();

        int stateSize() const { return ns->stateSize; }

        virtual void readState(const double* buf);
        virtual void writeState(double* out_buf) const;

        bool hasStateOnly() const { return (ns != 0); }

        const std::string& name() const { return ns->name; }
        void setName(const std::string& name) { ns->name = name; }
        const int id() const { return ns->id; }
        void setId(int id) { ns->id = id; }

        const Link* link() const { return ns->link; }
        Link* link() { return ns->link; }
        void setLink(Link* link) { ns->link = link; }

        Isometry3& T_local() { return ns->T_local; }
        const Isometry3& T_local() const { return ns->T_local; }
        
        Isometry3::ConstLinearPart R_local() const { return ns->const_T_local().linear(); }
        Isometry3::LinearPart R_local() { return ns->T_local.linear(); }

        Isometry3::ConstLinearPart localRotation() const { return ns->const_T_local().linear(); }
        Isometry3::LinearPart localRotaion() { return ns->T_local.linear(); }

        template<typename Derived>
        void setLocalRotation(const Eigen::MatrixBase<Derived>& R) { ns->T_local.linear() = R; }

        Isometry3::ConstTranslationPart p_local() const { return ns->const_T_local().translation(); }
        Isometry3::TranslationPart p_local() { return ns->T_local.translation(); }

        Isometry3::ConstTranslationPart localTranslation() const { return ns->const_T_local().translation(); }
        Isometry3::TranslationPart localTranslation() { return ns->T_local.translation(); }

        template<typename Derived>
        void setLocalTranslation(const Eigen::MatrixBase<Derived>& p) { ns->T_local.translation() = p; }
        
        double cycle() const { return ns->cycle; }
        void setCycle(double msec) { ns->cycle = msec; }

        SignalProxy<boost::signal<void()> > sigStateChanged() {
            return sigStateChanged_;
        }
        void notifyStateChange() const {
            sigStateChanged_();
        }
    };

    typedef boost::intrusive_ptr<Device> DevicePtr;


    template <class DeviceType = Device>
    class DeviceList : public PolymorphicPointerArray<DeviceType, boost::intrusive_ptr<DeviceType> >
    {
        typedef PolymorphicPointerArray<DeviceType, boost::intrusive_ptr<DeviceType> > ArrayBase;

    public:
        DeviceList() { }

        template <class RhsDeviceType>
        DeviceList(const DeviceList<RhsDeviceType>& rhs)
            : ArrayBase(rhs) { }

        template <class SubType>
        SubType* get(int index) const { return dynamic_cast<SubType*>(ArrayBase::operator[](index).get()); }

        DeviceType* get(int index) const { return ArrayBase::operator[](index).get(); }

        template<class TargetType> void makeIdMap(DeviceList<TargetType>& out_map) const {
            out_map.clear();
            for(size_t i=0; i < ArrayBase::size(); ++i){
                if(TargetType* device = dynamic_cast<TargetType*>(get(i))){
                    const int id = device->id();
                    if(id >= 0){
                        if(out_map.size() >= id){
                            out_map.resize(id + 1);
                        }
                        out_map[id] = device;
                    }
                }
            }
        }
    };

};

#endif
