/**
   @file
   @author Shin'ichiro Nakaoka
*/

#include "BodyMotionItem.h"
#include "BodyItem.h"
#include "KinematicFaultChecker.h"
#include <cnoid/MultiSeqItemCreationPanel>
#include <cnoid/BodyMotionUtil>
#include <cnoid/ItemManager>
#include <cnoid/Archive>
#include <cnoid/ZMPSeq>
#include <cnoid/LazyCaller>
#include <boost/bind.hpp>
#include <boost/format.hpp>
#include <QMessageBox>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

namespace {

    typedef boost::function<AbstractSeqItemPtr(AbstractSeqPtr seq)> ExtraSeqItemFactory;
    typedef map<string, ExtraSeqItemFactory> ExtraSeqItemFactoryMap;
    ExtraSeqItemFactoryMap extraSeqItemFactories;

    
    bool confirm(const std::string& message)
    {
        return (QMessageBox::warning(
                    0, _("Warning"), message.c_str(),
                    QMessageBox::Ok | QMessageBox::Cancel, QMessageBox::Ok) == QMessageBox::Ok);
    }

    bool fileIoSub(BodyMotionItem* item, std::ostream& os, bool loaded, bool isLoading)
    {
        if(!loaded){
            os << item->motion()->seqMessage();
        }
        return loaded;
    }
                
    bool loadStandardYamlFormat(BodyMotionItem* item, const std::string& filename, std::ostream& os)
    {
        return fileIoSub(item, os, item->motion()->loadStandardYAMLformat(filename), true);
    }
    
    bool saveAsStandardYamlFormat(BodyMotionItem* item, const std::string& filename, std::ostream& os)
    {
        return fileIoSub(item, os, item->motion()->saveAsStandardYAMLformat(filename), false);
    }

    bool importHrpsysSeqFileSet(BodyMotionItem* item, const std::string& filename, std::ostream& os)
    {
        if(loadHrpsysSeqFileSet(*item->motion(), filename, os)){
            return true;
        }
        return false;
    }
    
    
    bool exportHrpsysSeqFileSet(BodyMotionItem* item, const std::string& filename, std::ostream& os)
    {
        double frameRate = item->motion()->frameRate();
        if(frameRate != 200.0){
            static format m1(_("The frame rate of a body motion exported as HRPSYS files should be standard value 200, "
                               "but the frame rate of \"%1%\" is %2%. The exported data may cause a problem.\n\n"
                               "Do you continue to export ?"));

            if(!confirm(str(m1 % item->name() % frameRate))){
                return false;
            }
        }

        BodyPtr body;
        BodyItem* bodyItem = item->findOwnerItem<BodyItem>();
        if(bodyItem){
            body = bodyItem->body();
            KinematicFaultChecker* checker = KinematicFaultChecker::instance();
            int numFaults = checker->checkFaults(bodyItem, item, os);
            if(numFaults > 0){
                static string m2(_("A fault has been detected. Please check the report in the MessageView.\n\n"
                                   "Do you continue to export ?"));
                static format m3(_("%1% faults have been detected. Please check the report in the MessageView.\n\n"
                                   "Do you continue to export ?"));

                bool result;

                if(numFaults == 1){
                    result = confirm(m2);
                } else {
                    result = confirm(str(m3 % numFaults));
                }
                
                if(!result){
                    return false;
                }
            }
        }

        if(!getZMPSeq(*item->motion())){
            if(!confirm(_("There is no ZMP data. Do you continue to export ?"))){
                return false;
            }
        }
        
        return saveHrpsysSeqFileSet(*item->motion(), body, filename, os);
    }

    bool bodyMotionItemPreFilter(BodyMotionItem* protoItem, Item* parentItem)
    {
        BodyItemPtr bodyItem = dynamic_cast<BodyItem*>(parentItem);
        if(!bodyItem){
            bodyItem = parentItem->findOwnerItem<BodyItem>();
        }
        if(bodyItem){
            int prevNumJoints = protoItem->jointPosSeq()->numParts();
            int numJoints = bodyItem->body()->numJoints();
            if(numJoints != prevNumJoints){
                protoItem->jointPosSeq()->setNumParts(numJoints, true);
            }
        }
        return true;
    }


    /*
    bool bodyMotionItemPostFilter(BodyMotionItem* protoItem, Item* parentItem)
    {
        BodyItemPtr bodyItem = dynamic_cast<BodyItem*>(parentItem);
        if(!bodyItem){
            bodyItem = parentItem->findOwnerItem<BodyItem>();
        }
        if(bodyItem){
            BodyPtr body = bodyItem->body();
            MultiValueSeqPtr qseq = protoItem->jointPosSeq();
            int n = std::min(body->numJoints(), qseq->numParts());
            for(int i=0; i < n; ++i){
                Link* joint = body->joint(i);
                if(joint->defaultJointValue != 0.0){
                    MultiValueSeq::Part part = qseq->part(i);
                    std::fill(part.begin(), part.end(), joint->defaultJointValue);
                }
            }
        }
        
        return true;
    }
    */
}




void BodyMotionItem::initialize(ExtensionManager* ext)
{
    static bool initialized = false;
    
    if(initialized){
        return;
    }
    
    ItemManager& im = ext->itemManager();
    
    im.registerClass<BodyMotionItem>(N_("BodyMotionItem"));

    im.addCreationPanel<BodyMotionItem>(new MultiSeqItemCreationPanel(_("Number of joints")));
    im.addCreationPanelPreFilter<BodyMotionItem>(bodyMotionItemPreFilter);
    //im.addCreationPanelPostFilter<BodyMotionItem>(bodyMotionItemPostFilter);

    im.addLoaderAndSaver<BodyMotionItem>(
        _("Body Motion"), "BODY-MOTION-YAML", "yaml",
        bind(loadStandardYamlFormat, _1, _2, _3),  bind(saveAsStandardYamlFormat, _1, _2, _3));

    im.addLoaderAndSaver<BodyMotionItem>(
        _("HRPSYS Sequence File Set"), "HRPSYS-SEQ-FILE-SET", "pos;vel;acc;hip;waist;gsens;zmp",
        bind(importHrpsysSeqFileSet, _1, _2, _3), bind(exportHrpsysSeqFileSet, _1, _2, _3),
        ItemManager::PRIORITY_CONVERSION);

    initialized = true;
}


void BodyMotionItem::addExtraSeqItemFactory
(const std::string& key, boost::function<AbstractSeqItemPtr(AbstractSeqPtr seq)> factory)
{
    extraSeqItemFactories[key] = factory;
}


BodyMotionItem::BodyMotionItem()
    : bodyMotion_(new BodyMotion())
{
    initialize();
}


BodyMotionItem::BodyMotionItem(BodyMotionPtr bodyMotion)
    : bodyMotion_(bodyMotion)
{
    initialize();
}


BodyMotionItem::BodyMotionItem(const BodyMotionItem& org)
    : AbstractMultiSeqItem(org),
      bodyMotion_(new BodyMotion(*org.bodyMotion_))
{
    initialize();
}


BodyMotionItem::~BodyMotionItem()
{
    sigExtraSeqsChangedConnection.disconnect();
}


void BodyMotionItem::initialize()
{
    jointPosSeqItem_ = new MultiValueSeqItem(bodyMotion_->jointPosSeq());
    jointPosSeqItem_->setName("Joint");
    addSubItem(jointPosSeqItem_);

    jointPosSeqItem_->sigUpdated().connect(
        bind(&BodyMotionItem::onSubItemUpdated, this));

    linkPosSeqItem_ = new MultiSE3SeqItem(bodyMotion_->linkPosSeq());
    linkPosSeqItem_->setName("Cartesian");
    addSubItem(linkPosSeqItem_);

    linkPosSeqItem_->sigUpdated().connect(
        bind(&BodyMotionItem::onSubItemUpdated, this));

    sigExtraSeqsChangedConnection =
        bodyMotion_->sigExtraSeqsChanged().connect(
            bind(&BodyMotionItem::onExtraSeqItemSetChanged, this));

    updateExtraSeqItems();
}


AbstractMultiSeqPtr BodyMotionItem::abstractMultiSeq()
{
    return bodyMotion_;
}


void BodyMotionItem::onExtraSeqItemSetChanged()
{
    callLater(bind(&BodyMotionItem::updateExtraSeqItems, this));
}


void BodyMotionItem::updateExtraSeqItems()
{
    ExtraSeqItemMap prevItems(extraSeqItems);
    extraSeqItems.clear();
        
    BodyMotion::ConstSeqIterator p;
    for(p = bodyMotion_->extraSeqBegin(); p != bodyMotion_->extraSeqEnd(); ++p){
        const string& key = p->first;
        const AbstractSeqPtr& newSeq = p->second;
        AbstractSeqItemPtr newItem;
        ExtraSeqItemMap::iterator p = prevItems.find(key);
        if(p != prevItems.end()){
            ExtraSeqItemInfo& info = p->second;
            AbstractSeqItemPtr& prevItem = info.item;
            if(typeid(prevItem->abstractSeq()) == typeid(newSeq)){
                extraSeqItems.insert(make_pair(key, info));
                newItem = prevItem;
            } else {
                info.sigUpdateConnection.disconnect();
                prevItem->detachFromParentItem();
            }
            prevItems.erase(p);
        }
        if(!newItem){
            ExtraSeqItemFactoryMap::iterator q = extraSeqItemFactories.find(key);
            if(q != extraSeqItemFactories.end()){
                ExtraSeqItemFactory& factory = q->second;
                newItem = factory(newSeq);
                if(newItem){
                    addSubItem(newItem);
                    ExtraSeqItemInfo info(newItem);
                    info.sigUpdateConnection =
                        newItem->sigUpdated().connect(
                            bind(&BodyMotionItem::onSubItemUpdated, this));
                    extraSeqItems.insert(make_pair(key, info));
                }
            }
        }
    }
    for(ExtraSeqItemMap::iterator p = prevItems.begin(); p != prevItems.end(); ++p){
        ExtraSeqItemInfo& info = p->second;
        info.sigUpdateConnection.disconnect();
        info.item->detachFromParentItem();
    }

    sigExtraSeqItemsChanged_();
}


void BodyMotionItem::onSubItemUpdated()
{
    suggestFileUpdate();
    Item::notifyUpdate();
}


void BodyMotionItem::notifyUpdate()
{
    suggestFileUpdate();
    
    jointPosSeqItem_->notifyUpdate();
    linkPosSeqItem_->notifyUpdate();

    for(ExtraSeqItemMap::iterator p = extraSeqItems.begin(); p != extraSeqItems.end(); ++p){
        ExtraSeqItemInfo& info = p->second;
        info.item->notifyUpdate();
    }
    
    Item::notifyUpdate();
}


bool BodyMotionItem::onNewChildItem(Item* childItem, bool isManualOperation)
{
    if(isManualOperation){
        //! \todo copy the extra seq item dragged into this item, or create a new sub item from the dragged one.
    }
    return true;
}


ItemPtr BodyMotionItem::doDuplicate() const
{
    return new BodyMotionItem(*this);
}


bool BodyMotionItem::store(Archive& archive)
{
    if(overwrite()){
        archive.writeRelocatablePath("filename", lastAccessedFilePath());
        archive.write("format", lastAccessedFileFormatId());
        return true;
    }
    return false;
}


bool BodyMotionItem::restore(const Archive& archive)
{
    std::string filename, formatId;
    if(archive.readRelocatablePath("filename", filename) && archive.read("format", formatId)){
        if(load(filename, formatId)){
            return true;
        }
    }
    return false;
}
