/**
   @author Shin'ichiro Nakaoka
*/

#include "SceneBody.h"
#include "BodyItem.h"
#include "KinematicsBar.h"
#include "BodyBar.h"
#include "LinkSelectionView.h"
#include <cnoid/Link>
#include <cnoid/Camera>
#include <cnoid/Light>
#include <cnoid/ConnectionSet>
#include <cnoid/JointPath>
#include <cnoid/PenetrationBlocker>
#include <cnoid/MessageView>
#include <cnoid/MenuManager>
#include <cnoid/SceneView>
#include <cnoid/ScenePieces>
#include <cnoid/GLSceneRenderer>
#include <cnoid/LeggedBodyHelper>
#include <cnoid/ColdetLinkPair>
#include <cnoid/PinDragIK>
#include <cnoid/EigenUtil>
#include <GL/glew.h>
#include <boost/bind.hpp>
#include <iostream>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

namespace {
    const bool TRACE_FUNCTIONS = false;

    struct compare {
        bool operator ()(const std::type_info* a, const std::type_info* b) const {
            return a->before(*b);
        }
    };

    typedef std::map<const std::type_info*, SceneBody::DeviceNodeFactory, compare> DeviceNodeFactoryMap;
    DeviceNodeFactoryMap deviceNodeFactories;
    
    class SceneLink : public SgTransform
    {
    public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW

        const Link* link;
        SgNodePtr shape;
        SgGroupPtr devices;
        SgCustomGLNodePtr marker;
        SgBBMarkerPtr bbMarker;
        bool isPointed;
        bool isColliding;

        SceneLink(const Link* link) : link(link) {
            setName(link->name());
            shape = link->shape();
            if(shape){
                addChild(shape);
            }
            bbMarker = 0;
        }

        void updatePosition() {
            setTranslation(link->p().cast<float>());
            setRotation(link->attitude());
        }
        
        void addSceneDevice(SceneDevice* sdev) {
            if(!devices){
                devices = new SgGroup();
                addChild(devices);
            }
            devices->addChild(sdev);
        }

        double calcMarkerRadius() const {
            const SgBoundingBox& bb = shape->boundingBox();
            double V = ((bb.max().x() - bb.min().x()) * (bb.max().y() - bb.min().y()) * (bb.max().z() - bb.min().z()));
            return pow(V, 1.0 / 3.0) * 0.6;
        }

        void showBoundingBox(bool on) {
            if(!shape){
                return;
            }
            if(on){
                if(!marker){
                    marker = new SgCustomGLNode(bind(&SceneLink::renderMarker, this, _1));
                }
                if(!contains(marker)){
                    addChild(marker);
                    notifyUpdate();
                }
            } else if(marker){
                if(removeChild(marker)){
                    notifyUpdate();
                }
            }
        }
        void renderMarker(GLSceneRenderer& renderer) {
            
            glPushAttrib(GL_LIGHTING_BIT | GL_CURRENT_BIT);
            glDisable(GL_LIGHTING);

            //! \todo use GLSceneRenderer's state change commands
            glColor3f(1.0f, 0.0f, 0.0f);

            glBegin(GL_LINES);

            const SgBoundingBox& bb = shape->boundingBox();
            const SgVector3& min = bb.min();
            const SgVector3& max = bb.max();

            SgFloat y[2];
            y[0] = min.y();
            y[1] = max.y();
            
            for(int i=0; i < 2; ++i){
                glVertex3d(min.x(), y[i], min.z());
                glVertex3d(min.x(), y[i], max.z());

                glVertex3d(min.x(), y[i], max.z());
                glVertex3d(max.x(), y[i], max.z());

                glVertex3d(max.x(), y[i], max.z());
                glVertex3d(max.x(), y[i], min.z());

                glVertex3d(max.x(), y[i], min.z());
                glVertex3d(min.x(), y[i], min.z());
            }

            SgFloat z[2];
            z[0] = min.z();
            z[1] = max.z();
            for(int i=0; i < 2; ++i){
                glVertex3d(min.x(), min.y(), z[i]);
                glVertex3d(min.x(), max.y(), z[i]);

                glVertex3d(max.x(), min.y(), z[i]);
                glVertex3d(max.x(), max.y(), z[i]);
            }
            
            glEnd();
            glPopAttrib();
        }

        void showMarker(const Vector4& color, float width)
        {
            if(!bbMarker){
                const SgBoundingBox& bb = shape->boundingBox();
                bbMarker = new SgBBMarker(bb, color);
            }
            if(!contains(bbMarker)){
                addChild(bbMarker);
                notifyUpdate();
            }
        }

        void hideMarker()
        {
            if(bbMarker){
                if(removeChild(bbMarker)){
                    notifyUpdate();
                }
            }
        }

    };
    typedef boost::intrusive_ptr<SceneLink> SceneLinkPtr;

    class SgCollision : public SgCustomGLNode
    {
        const std::vector<ColdetLinkPairPtr>* ppairs;
        
    public:
        SgCollision()
        {
            ppairs = 0;
        }

        void setColdetPairs(const std::vector<ColdetLinkPairPtr>& pairs) {
            ppairs = &pairs;
        };

        virtual void render(GLSceneRenderer& renderer)
        {
            if(!renderer.property()->get("collision", false)){
                return;
            }
            
            glPushAttrib(GL_LIGHTING_BIT | GL_CURRENT_BIT);
            glDisable(GL_LIGHTING);
            glColor3f(0.0f, 1.0f, 0.0f);

            glBegin(GL_LINES);

            const std::vector<ColdetLinkPairPtr>& pairs = *ppairs;

            for(size_t i=0; i < pairs.size(); ++i){
                const std::vector<collision_data>& cols = pairs[i]->collisions();
                for(size_t j=0; j < cols.size(); ++j){
                    const collision_data& col = cols[j];
                    const Vector3 n = 50.0 * col.depth * col.n_vector;
                    for(int k=0; k < col.num_of_i_points; ++k){
                        if(col.i_point_new[k]){
                            const Vector3& p = col.i_points[k];
                            glVertex3dv(p.data());
                            glVertex3dv(Vector3(p + n).data());
                        }
                    }
                }
            }

            glEnd();
            glPopAttrib();
        }
    };
    typedef boost::intrusive_ptr<SgCollision> SgCollisionPtr;

}


namespace  cnoid {

    class SceneBodyImpl
    {
    public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW

        SceneBodyImpl(SceneBody* self, BodyItemPtr& bodyItem);
        ~SceneBodyImpl();
        
        SceneBody* self;
        BodyItemPtr bodyItem;
        BodyPtr body;
        vector<SceneLinkPtr> sceneLinks;
        vector<SceneDevicePtr> sceneDevices;
        ConnectionSet connections;
        signals::connection connectionToSigWorldCollisionLinkSetChanged;

        enum PointedType { PT_NONE, PT_SCENE_LINK, PT_ZMP };
        SceneLink* pointedSceneLink;
        SceneLink* outlinedLink;

        SgCollisionPtr sgCollision; 
        SgGroupPtr markerGroup;
        SgCrossMarkerPtr cmMarker;
        bool isCmVisible;
        SgSphereMarkerPtr zmpMarker;
        bool isZmpVisible;
        Vector3 orgZmpPos;

        Link* targetLink;
        Vector3 orgPointerPos;
        Vector3 orgBaseLinkPos;
        Vector3 orgTargetLinkPos;
        double orgJointPosition;
        Vector3 rotationBaseX;
        Vector3 rotationBaseY;
        Matrix3 orgAttitude;
        Vector3 axis;
        
        JointPathPtr ikPath;
        LinkTraverse fkTraverse;
        PinDragIKptr pinDragIK;
        InverseKinematicsPtr ik;
        PenetrationBlockerPtr penetrationBlocker;
        SgAttitudeDraggerPtr attitudeDragger;

        SceneViewEditable::SceneMode sceneMode;

        KinematicsBar* kinematicsBar;

        enum Axis { NOAXIS = -1, X = 0, Y = 1, Z = 2 };
        Axis rotationAxis;
        enum DragMode {
            DRAG_NONE,
            LINK_IK_TRANSLATION, LINK_IK_ROTATION,
            LINK_FK_ROTATION, LINK_FK_TRANSLATION,
            ZMP_TRANSLATION
        };
        DragMode dragMode;
        bool isDragging;

        shared_ptr<SceneProjector> projector;

        bool setSceneDevice(const std::type_info& type, Device* device);
        void onSceneGraphConnection(bool on);
        void onBodyItemUpdated();
        void onKinematicStateChanged();

        static SgNode* createPerspectiveCamera(SceneDevice* sdev);
        static SgNode* createPointLight(SceneDevice* sdev);
        static SgNode* createSpotLight(SceneDevice* sdev);

        void updatePositions(bool doNotify);

        bool setWorldCollisionLinkVisual(SceneLink* sceneLink, bool on);
        bool setSelfCollisionLinkVisual(SceneLink* sceneLink, bool on);
        //void onSelfCollisionsUpdated();
        void onWorldCollisionLinkSetChanged();
        void onCollisionLinkHighlightModeChanged();
        void changeCollisionLinkHighlightMode(bool on);

        void showCenterOfMass(bool on);
        void showZmp(bool on);
        void makeLinkFree(SceneLink* sceneLink);
        void setBaseLink(SceneLink* sceneLink);
        void toggleBaseLink(SceneLink* sceneLink);
        void togglePin(SceneLink* sceneLink, bool toggleTranslation, bool toggleRotation);
        void makeLinkAttitudeLevel();
        
        SceneBodyImpl::PointedType findPointedObject(const vector<SgNode*>& path);
        void updateMarkersAndManipulators();

        bool onKeyPressEvent(const SceneViewEvent& event);
        bool onKeyReleaseEvent(const SceneViewEvent& event);
        bool onButtonPressEvent(const SceneViewEvent& event);
        bool onDoubleClickEvent(const SceneViewEvent& event);
        bool onButtonReleaseEvent(const SceneViewEvent& event);
        bool onPointerMoveEvent(const SceneViewEvent& event);
        void onPointerLeaveEvent(const SceneViewEvent& event);
        bool onScrollEvent(const SceneViewEvent& event);
        void onContextMenuRequest(const SceneViewEvent& event, MenuManager& menuManager);
        void onSceneModeChanged(SceneViewEditable::SceneMode mode);
        bool onUndoRequest();
        bool onRedoRequest();

        void attachAttitudeDraggerTo(Link* link);
        void startIK(const SceneViewEvent& event);
        void dragIK(const SceneViewEvent& event);
        void startFK(const SceneViewEvent& event);
        void dragFKRotation(const SceneViewEvent& event);
        void dragFKTranslation(const SceneViewEvent& event);
        void startZmpTranslation(const SceneViewEvent& event);
        void dragZmpTranslation(const SceneViewEvent& event);
    };
}
    
    
namespace {

    struct DeviceNodeFactoryMapInitializer
    {
        DeviceNodeFactoryMapInitializer() {
            deviceNodeFactories[&typeid(Camera)] = &SceneBodyImpl::createPerspectiveCamera;
            deviceNodeFactories[&typeid(PointLight)] = &SceneBodyImpl::createPointLight;
            deviceNodeFactories[&typeid(SpotLight)]  = &SceneBodyImpl::createSpotLight;
        }
    };
    DeviceNodeFactoryMapInitializer initializer;
}


void SceneBody::registerDeviceNodeFactory_(const std::type_info* pTypeInfo, const DeviceNodeFactory& factory)
{
    deviceNodeFactories[pTypeInfo] = factory;
}


SceneDevice::SceneDevice(Device* device)
  : device_(device)
{
    connections = new ConnectionSet;
}


SceneDevice::~SceneDevice()
{
    connections->disconnect();
    delete connections;
}


void SceneDevice::addUpdateConnection(const boost::signals::connection& connection)
{
    connections->add(connection);
}


SceneBody::SceneBody(BodyItemPtr bodyItem)
{
    impl = new SceneBodyImpl(this, bodyItem);
}


SceneBodyImpl::SceneBodyImpl(SceneBody* self, BodyItemPtr& bodyItem)
    : self(self),
      bodyItem(bodyItem),
      body(bodyItem->body()),
      kinematicsBar(KinematicsBar::instance())
{
    const BodyPtr& body = bodyItem->body();

    self->setName(body->name());
    
    double radius = 0;
    const int n = body->numLinks();
    for(int i=0; i < n; ++i){
        const Link* link = body->link(i);
        SceneLink* sLink = new SceneLink(link);
        self->addChild(sLink);
        sceneLinks.push_back(sLink);
        SgBoundingBox bb = sLink->boundingBox();
        double radius0 = (bb.max() - bb.center()).norm();
        if( radius0 > radius )
            radius = radius0;
    }

    const DeviceList<Device>& devices = body->devices();
    for(size_t i=0; i < devices.size(); ++i){
        Device* device = devices.get(i);
        device->forEachActualType(bind(&SceneBodyImpl::setSceneDevice, this, _1, device));
    }

    updatePositions(false);

    pointedSceneLink = 0;
    outlinedLink = 0;
    targetLink = 0;
    attitudeDragger = new SgAttitudeDragger();

    dragMode = DRAG_NONE;
    isDragging = false;

    self->setEditable(!body->isStaticModel());

    sgCollision = new SgCollision();
    sgCollision->setColdetPairs(bodyItem->selfColdetPairs());
    self->addChild(sgCollision);
    connections.add(kinematicsBar->sigCollisionVisualizationChanged().connect(
        bind(&SceneBodyImpl::onCollisionLinkHighlightModeChanged, this)));

    markerGroup = new SgGroup;
    self->addChild(markerGroup);

    cmMarker = new SgCrossMarker(Vector4(0.0, 1.0, 0.0, 1.0), 0.25f, 2.0f);
    cmMarker->setSize(radius);
    isCmVisible = false;

     LeggedBodyHelperPtr legged = getLeggedBodyHelper(body);
     if(legged->isValid() && legged->numFeet() > 0){
        Link* footLink = legged->footLink(0);
        SceneLinkPtr sceneLink = sceneLinks[footLink->index()];
        double r = sceneLink->calcMarkerRadius();
        zmpMarker = new SgSphereMarker(r, Vector4(0.0f, 1.0f, 0.0f, 0.3f));
        zmpMarker->setCross(Vector4(0.0, 1.0, 0.0, 1.0), r * 2.5, 2.0f);
    }else{
        zmpMarker = new SgSphereMarker(0.1, Vector4(0.0f, 1.0f, 0.0f, 0.3f));
        zmpMarker->setCross(Vector4(0.0, 1.0, 0.0, 1.0), 0.25f, 2.0f);
     }
    isZmpVisible = false;

    self->sigGraphConnection().connect(bind(&SceneBodyImpl::onSceneGraphConnection, this, _1));

    onCollisionLinkHighlightModeChanged();

    updateMarkersAndManipulators();
    onKinematicStateChanged();
}


void SceneBodyImpl::onSceneGraphConnection(bool on)
{
    connections.disconnect();

    if(on){
        connections.add(bodyItem->sigUpdated().connect(
                            bind(&SceneBodyImpl::onBodyItemUpdated, this)));
        connections.add(bodyItem->sigKinematicStateChanged().connect(
                            bind(&SceneBodyImpl::onKinematicStateChanged, this)));
        //connections.add(bodyItem->sigSelfCollisionsUpdated().connect(
        //    bind(&SceneBodyImpl::onSelfCollisionsUpdated, this)));
    }
}


void SceneBodyImpl::onBodyItemUpdated()
{
    updateMarkersAndManipulators();
}


void SceneBodyImpl::onKinematicStateChanged()
{
    if(TRACE_FUNCTIONS){
        static int counter = 0;
        cout << "SceneBodyImpl::onKinematicStateChanged():" << counter++ << endl;
    }
    
    if(isCmVisible){
        cmMarker->setPosition(bodyItem->centerOfMass());
    }
    if(isZmpVisible){
        zmpMarker->setPosition(bodyItem->zmp());
    }

    updatePositions(true);
}


bool SceneBodyImpl::setWorldCollisionLinkVisual(SceneLink* sceneLink, bool on)
{
    if(!sceneLink->isColliding && on){
        if(!sceneLink->isPointed){
        }
        sceneLink->isColliding = true;
    } else if(sceneLink->isColliding && !on){
        if(!sceneLink->isPointed){
        }
        sceneLink->isColliding = false;
    }

    return true;
}

bool SceneBodyImpl::setSelfCollisionLinkVisual(SceneLink* sceneLink, bool on)
{
    if(on){

    } else {

    }
    return true;
}


//void SceneBodyImpl::onSelfCollisionsUpdated()
//{
//
//}

void SceneBodyImpl::onWorldCollisionLinkSetChanged()
{
    const boost::dynamic_bitset<>& w = bodyItem->worldCollisionLinkBitSet();
    const boost::dynamic_bitset<>& s = bodyItem->selfCollisionLinkBitSet();
    for(size_t i=0; i < sceneLinks.size(); ++i){
        setWorldCollisionLinkVisual(sceneLinks[i].get(), (w[i] && !s[i]));
        setSelfCollisionLinkVisual(sceneLinks[i].get(), s[i]);
    }

    self->notifyUpdate();
}


void SceneBodyImpl::onCollisionLinkHighlightModeChanged()
{
    changeCollisionLinkHighlightMode(kinematicsBar->isCollisionLinkHighlihtMode());
}


void SceneBodyImpl::changeCollisionLinkHighlightMode(bool on)
{
    if(!connectionToSigWorldCollisionLinkSetChanged.connected() && on){

        connectionToSigWorldCollisionLinkSetChanged =
            bodyItem->sigWorldCollisionLinkSetChanged().connect(
                bind(&SceneBodyImpl::onWorldCollisionLinkSetChanged, this));
        onWorldCollisionLinkSetChanged();

    } else if(connectionToSigWorldCollisionLinkSetChanged.connected() && !on){

        connectionToSigWorldCollisionLinkSetChanged.disconnect();
        for(size_t i=0; i < sceneLinks.size(); ++i){
            setWorldCollisionLinkVisual(sceneLinks[i].get(), false);
            setSelfCollisionLinkVisual(sceneLinks[i].get(), false);
        }
        self->notifyUpdate();
    }
}


SceneBody::~SceneBody()
{
    delete impl;
}


SceneBodyImpl::~SceneBodyImpl()
{
    connectionToSigWorldCollisionLinkSetChanged.disconnect();
    connections.disconnect();
}


bool SceneBodyImpl::setSceneDevice(const std::type_info& type, Device* device)
{
    DeviceNodeFactoryMap::iterator p = deviceNodeFactories.find(&type);

    if(p != deviceNodeFactories.end()){

        SceneBody::DeviceNodeFactory& factory = p->second;
        SceneDevice* sceneDevice = new SceneDevice(device);
        SgNode* node = factory(sceneDevice);
        node->setName(device->name());
        sceneDevice->addChild(node);
        sceneDevice->setTransform(device->T_local());

        sceneLinks[device->link()->index()]->addSceneDevice(sceneDevice);
        sceneDevices.push_back(sceneDevice);
        
        return true;
    }
    
    return false;
}


SgNode* SceneBodyImpl::createPerspectiveCamera(SceneDevice* sdev)
{
    Camera* camera = sdev->device<Camera>();
    SgPerspectiveCamera* scamera = new SgPerspectiveCamera();
    sdev->addUpdateConnection(
        camera->sigStateChanged().connect(
            bind(&SceneBody::updatePerspectiveCamera, camera, scamera)));
    SceneBody::updatePerspectiveCamera(camera, scamera);
    return scamera;
}


void SceneBody::updatePerspectiveCamera(Camera* camera, SgPerspectiveCamera* scamera)
{
    scamera->setFieldOfView(camera->fieldOfView());
}


void SceneBody::updateLight(Light* light, SgLight* slight)
{
    slight->on(light->on());
    slight->setColor(light->color());
    slight->setIntensity(light->intensity());
}


SgNode* SceneBodyImpl::createPointLight(SceneDevice* sdev)
{
    PointLight* light = sdev->device<PointLight>();
    SgPointLight* slight = new SgPointLight();
    sdev->addUpdateConnection(
        light->sigStateChanged().connect(
            bind(&SceneBody::updatePointLight, light, slight)));
    SceneBody::updatePointLight(light, slight);
    return slight;
}


void SceneBody::updatePointLight(PointLight* light, SgPointLight* slight)
{
    SceneBody::updateLight(light, slight);
    slight->setAttenuation(light->attenuation());
}


SgNode* SceneBodyImpl::createSpotLight(SceneDevice* sdev)
{
    SpotLight* light = sdev->device<SpotLight>();
    SgSpotLight* slight = new SgSpotLight();
    sdev->addUpdateConnection(
        light->sigStateChanged().connect(
            bind(&SceneBody::updateSpotLight, light, slight)));
    SceneBody::updateSpotLight(light, slight);
    return slight;
}


void SceneBody::updateSpotLight(SpotLight* light, SgSpotLight* slight)
{
    SceneBody::updatePointLight(light, slight);

    slight->setDirection(light->direction());
    slight->setBeamWidth(light->beamWidth());
    slight->setCutOffAngle(light->cutOffAngle());
}


void SceneBody::updatePositions(bool doNotify)
{
    impl->updatePositions(doNotify);
}


void SceneBodyImpl::updatePositions(bool doNotify)
{
    for(size_t i=0; i < sceneLinks.size(); ++i){
        sceneLinks[i]->updatePosition();
    }
    if(doNotify){
        self->notifyUpdate();
    }
}


void SceneBodyImpl::showCenterOfMass(bool on)
{
    isCmVisible = on;
    if(on){
        markerGroup->addChild(cmMarker);
        cmMarker->setPosition(bodyItem->centerOfMass());
    } else {
        markerGroup->removeChild(cmMarker);
    }
    self->notifyUpdate();
}


void SceneBodyImpl::showZmp(bool on)
{
    isZmpVisible = on;
    if(on){
        markerGroup->addChild(zmpMarker);
        zmpMarker->setPosition(bodyItem->zmp());
    } else {
        markerGroup->removeChild(zmpMarker);
    }
    self->notifyUpdate();
}


void SceneBodyImpl::makeLinkFree(SceneLink* sceneLink)
{
    Link* link = (Link*)sceneLink->link;
    if(bodyItem->currentBaseLink() == link){
        bodyItem->setCurrentBaseLink(0);
    }
    
    bodyItem->pinDragIK()->setPin(link, InverseKinematics::NO_AXES);

    bodyItem->notifyUpdate();
}


void SceneBodyImpl::setBaseLink(SceneLink* sceneLink)
{
    bodyItem->setCurrentBaseLink((Link*)sceneLink->link);
    bodyItem->notifyUpdate();
}
    

void SceneBodyImpl::toggleBaseLink(SceneLink* sceneLink)
{
    Link* baseLink = bodyItem->currentBaseLink();
    if(sceneLink->link != baseLink){
        bodyItem->setCurrentBaseLink((Link*)sceneLink->link);
    } else {
        bodyItem->setCurrentBaseLink(0);
    }
    bodyItem->notifyUpdate();
}


void SceneBodyImpl::togglePin(SceneLink* sceneLink, bool toggleTranslation, bool toggleRotation)
{
    Link* link = (Link*)sceneLink->link;
    PinDragIKptr pin = bodyItem->pinDragIK();

    InverseKinematics::AxisSet axes = pin->pinAxes(link);

    if(toggleTranslation && toggleRotation){
        if(axes == InverseKinematics::NO_AXES){
            axes = InverseKinematics::TRANSFORM_6D;
        } else {
            axes = InverseKinematics::NO_AXES;
        }
    } else {
        if(toggleTranslation){
            axes = (InverseKinematics::AxisSet)(axes ^ InverseKinematics::TRANSLATION_3D);
        }
        if(toggleRotation){
            axes = (InverseKinematics::AxisSet)(axes ^ InverseKinematics::ROTATION_3D);
        }
    }
        
    pin->setPin(link, axes);
    bodyItem->notifyUpdate();
}


void SceneBodyImpl::makeLinkAttitudeLevel()
{
    if(pointedSceneLink){
        Link* targetLink = (Link*)outlinedLink->link;
        InverseKinematicsPtr ik = bodyItem->getCurrentIK(targetLink);
        if(ik){
            const Position& T = targetLink->T();
            const double theta = acos(T(2, 2));
            const Vector3 z(T(0,2), T(1, 2), T(2, 2));
            const Vector3 axis = z.cross(Vector3::UnitZ()).normalized();
            const Matrix3 R2 = AngleAxisd(theta, axis) * T.linear();

            bodyItem->beginKinematicStateEdit();
            if(ik->calcInverseKinematics(targetLink->p(), R2)){
                bodyItem->notifyKinematicStateChange(true);
                bodyItem->acceptKinematicStateEdit();
            }
        }
    }
}


void SceneBodyImpl::updateMarkersAndManipulators()
{
    bool show = (sceneMode == SceneViewEditable::EDIT_MODE && self->isEditable());
    
    Link* baseLink = bodyItem->currentBaseLink();
    PinDragIKptr pin = bodyItem->pinDragIK();
    
    for(size_t i=0; i < sceneLinks.size(); ++i){
        SceneLink* sceneLink = sceneLinks[i].get();
        Link* link = (Link*)sceneLink->link;
        sceneLink->hideMarker();
        if(show){
            if(link == baseLink){
                sceneLink->showMarker(Vector4(1.0f, 0.1f, 0.1f, 0.4f), 0.02);
            } else {
                int pinAxes = pin->pinAxes(link);
                if(pinAxes & (InverseKinematics::TRANSFORM_6D)){
                    sceneLink->showMarker(Vector4(1.0f, 1.0f, 0.1f, 0.4f), 0.02);
                }
            }
        }
    }

    attitudeDragger->detach();

    if(show && targetLink && (kinematicsBar->mode() == KinematicsBar::IK_MODE) && kinematicsBar->isAttitudeMode()){
        attitudeDragger->attachTo(sceneLinks[targetLink->index()]);
    }

    self->notifyUpdate();
}


SceneBodyImpl::PointedType SceneBodyImpl::findPointedObject(const vector<SgNode*>& path)
{
    PointedType pointedType = PT_NONE;
    pointedSceneLink = 0;
    rotationAxis = NOAXIS;
    for(size_t i = path.size() - 1; i >= 1; --i){
        pointedSceneLink = dynamic_cast<SceneLink*>(path[i]);
        if(pointedSceneLink){
            pointedType = PT_SCENE_LINK;
            if(i < path.size() - 2){
                SgAttitudeDragger* dragger = dynamic_cast<SgAttitudeDragger*>(path[i+1]);
                if(dragger){
                    const string& axis = path[i+2]->name();
                    if(axis == "x"){
                        rotationAxis = X;
                    } else if(axis == "y"){
                        rotationAxis = Y;
                    } else if(axis == "z"){
                        rotationAxis = Z;
                    }
                }
            }
            break;
        }
        SgSphereMarker* marker = dynamic_cast<SgSphereMarker*>(path[i]);
        if(marker == zmpMarker){
            pointedType = PT_ZMP;
            break;
        }
    }
    return pointedType;
}


bool SceneBody::onKeyPressEvent(const SceneViewEvent& event)
{
    return impl->onKeyPressEvent(event);
}


bool SceneBodyImpl::onKeyPressEvent(const SceneViewEvent& event)
{
    if(!outlinedLink){
        return false;
    }

    bool handled = true;

    switch(event.key()){
    case Qt::Key_B:
        toggleBaseLink(outlinedLink);
        break;
        
    case Qt::Key_R:
        togglePin(outlinedLink, false, true);
        break;

    case Qt::Key_T:
        togglePin(outlinedLink, true, false);
        break;

    default:
        handled = false;
        break;
    }
        
    return handled;

}


bool SceneBody::onKeyReleaseEvent(const SceneViewEvent& event)
{
    return impl->onKeyReleaseEvent(event);
}


bool SceneBodyImpl::onKeyReleaseEvent(const SceneViewEvent& event)
{
    return false;
}


bool SceneBody::onButtonPressEvent(const SceneViewEvent& event)
{
    return impl->onButtonPressEvent(event);
}


bool SceneBodyImpl::onButtonPressEvent(const SceneViewEvent& event)
{
    if(TRACE_FUNCTIONS){
        cout << "SceneBodyImpl::onButtonPressEvent()" << endl;
    }
    
    if(outlinedLink){
        outlinedLink->showBoundingBox(false);
        outlinedLink = 0;
    }

    PointedType pointedType = findPointedObject(event.nodePath());
    
    if(pointedSceneLink){
        pointedSceneLink->showBoundingBox(true);
        outlinedLink = pointedSceneLink;
    }
    
    bool handled = false;
    isDragging = false;

    if(pointedType == PT_SCENE_LINK){

        if(event.button() == Qt::LeftButton){
            targetLink = (Link *)pointedSceneLink->link;
            updateMarkersAndManipulators();
            ik.reset();

            switch(kinematicsBar->mode()){
            case KinematicsBar::AUTO_MODE:
                ik = bodyItem->getDefaultIK(targetLink);
                if(ik){
                    attitudeDragger->attachTo(sceneLinks[targetLink->index()]);
                    startIK(event);
                } else {
                    startFK(event);
                }
                break;
            case KinematicsBar::FK_MODE:
                if(targetLink == bodyItem->currentBaseLink()){
                    startIK(event);
                } else {
                    startFK(event);
                }
                break;
            case KinematicsBar::IK_MODE:
                startIK(event);
                break;
            }
            
        } else if(event.button() == Qt::MiddleButton){
            togglePin(pointedSceneLink, true, true);
            
        } else if(event.button() == Qt::RightButton){

        }

        handled = true;

    } else if(pointedType == PT_ZMP){
        startZmpTranslation(event);
        handled = true;
    }

    if(dragMode != DRAG_NONE && outlinedLink){
        outlinedLink->showBoundingBox(false);
        self->notifyUpdate();
    }

    return handled;
}


bool SceneBody::onButtonReleaseEvent(const SceneViewEvent& event)
{
    return impl->onButtonReleaseEvent(event);
}


bool SceneBodyImpl::onButtonReleaseEvent(const SceneViewEvent& event)
{
    isDragging = false;
    
    if(dragMode != DRAG_NONE){
        bodyItem->acceptKinematicStateEdit();
        dragMode = DRAG_NONE;
        if(outlinedLink){
            outlinedLink->showBoundingBox(true);
            self->notifyUpdate();
        }
        return true;
    }
    return false;
}


bool SceneBody::onDoubleClickEvent(const SceneViewEvent& event)
{
    return impl->onDoubleClickEvent(event);
}


bool SceneBodyImpl::onDoubleClickEvent(const SceneViewEvent& event)
{
    if(findPointedObject(event.nodePath()) == PT_SCENE_LINK){
        if(event.button() == Qt::LeftButton){
            if(BodyBar::instance()->makeSingleSelection(bodyItem)){
                LinkSelectionView::mainInstance()->makeSingleSelection(bodyItem, pointedSceneLink->link->index());
            }
            return true;
        }
    }
    return false;
}


bool SceneBody::onPointerMoveEvent(const SceneViewEvent& event)
{
    return impl->onPointerMoveEvent(event);
}


bool SceneBodyImpl::onPointerMoveEvent(const SceneViewEvent& event)
{
    if(dragMode == DRAG_NONE){
        PointedType pointedType = findPointedObject(event.nodePath());
        if(pointedSceneLink){
            if(pointedSceneLink != outlinedLink){
                if(outlinedLink)
                    outlinedLink->showBoundingBox(false);
                pointedSceneLink->showBoundingBox(true);
                outlinedLink = pointedSceneLink;
            }
        }
        
        static format f(_("%1% / %2%"));
        if(pointedSceneLink){
            event.updateIndicator(str(f % bodyItem->name() % pointedSceneLink->link->name()));
        } else {
            event.updateIndicator("");
        }
    } else {
        if(!isDragging){
            bodyItem->beginKinematicStateEdit();
            isDragging = true;
        }

        switch(dragMode){

        case LINK_IK_TRANSLATION:
        case LINK_IK_ROTATION:
            dragIK(event);
            break;
            
        case LINK_FK_ROTATION:
            dragFKRotation(event);
            break;

        case LINK_FK_TRANSLATION:
            dragFKTranslation(event);
            break;
            
        case ZMP_TRANSLATION:
            dragZmpTranslation(event);
            break;
            
        default:
            break;
        }
    }

    return true;
}


void SceneBody::onPointerLeaveEvent(const SceneViewEvent& event)
{
    return impl->onPointerLeaveEvent(event);
}


void SceneBodyImpl::onPointerLeaveEvent(const SceneViewEvent& event)
{
    if(outlinedLink){
        outlinedLink->showBoundingBox(false);
        outlinedLink = 0;
    }
}


bool SceneBody::onScrollEvent(const SceneViewEvent& event)
{
    return impl->onScrollEvent(event);
}


bool SceneBodyImpl::onScrollEvent(const SceneViewEvent& event)
{
    return false;
}


void SceneBody::onContextMenuRequest(const SceneViewEvent& event, MenuManager& menuManager)
{
    impl->onContextMenuRequest(event, menuManager);
}


void SceneBodyImpl::onContextMenuRequest(const SceneViewEvent& event, MenuManager& menuManager)
{
    PointedType pointedType = findPointedObject(event.nodePath());

    if(bodyItem && pointedType == PT_SCENE_LINK){

        menuManager.addItem(_("Set Free"))->sigTriggered().connect(
            bind(&SceneBodyImpl::makeLinkFree, this, pointedSceneLink));
        menuManager.addItem(_("Set Base"))->sigTriggered().connect(
            bind(&SceneBodyImpl::setBaseLink, this, pointedSceneLink));
        menuManager.addItem(_("Set Translation Pin"))->sigTriggered().connect(
            bind(&SceneBodyImpl::togglePin, this, pointedSceneLink, true, false));
        menuManager.addItem(_("Set Rotation Pin"))->sigTriggered().connect(
            bind(&SceneBodyImpl::togglePin, this, pointedSceneLink, false, true));
        menuManager.addItem(_("Set Both Pins"))->sigTriggered().connect(
            bind(&SceneBodyImpl::togglePin, this, pointedSceneLink, true, true));

        menuManager.addSeparator();

        menuManager.addItem(_("Level Attitude"))->sigTriggered().connect(
            bind(&SceneBodyImpl::makeLinkAttitudeLevel, this));

        menuManager.addSeparator();
        
        menuManager.setPath(N_("/Markers"));
        
        Action* item = menuManager.addCheckItem(_("Center of Mass"));
        item->setChecked(isCmVisible);
        item->sigToggled().connect(bind(&SceneBodyImpl::showCenterOfMass, this, _1));

        item = menuManager.addCheckItem(_("ZMP"));
        item->setChecked(isZmpVisible);
        item->sigToggled().connect(bind(&SceneBodyImpl::showZmp, this, _1));
    }
}


void SceneBody::onSceneModeChanged(SceneMode mode)
{
    impl->onSceneModeChanged(mode);
}


void SceneBodyImpl::onSceneModeChanged(SceneViewEditable::SceneMode mode)
{
    sceneMode = mode;

    if(outlinedLink){
        if(sceneMode == SceneViewEditable::EDIT_MODE){
            outlinedLink->showBoundingBox(true);
        } else {
            outlinedLink->showBoundingBox(false);
            outlinedLink = 0;
        }
    }
    updateMarkersAndManipulators();
}


bool SceneBody::onUndoRequest()
{
    return impl->onUndoRequest();
}


bool SceneBodyImpl::onUndoRequest()
{
    return bodyItem->undoKinematicState();
}


bool SceneBody::onRedoRequest()
{
    return impl->onRedoRequest();
}


bool SceneBodyImpl::onRedoRequest()
{
    return bodyItem->redoKinematicState();
}


void SceneBodyImpl::startIK(const SceneViewEvent& event)
{
    orgPointerPos << event.point().x(), event.point().y(), event.point().z();
    orgTargetLinkPos = targetLink->p();

    Link* baseLink = bodyItem->currentBaseLink();

    if(!ik){
        if(bodyItem->pinDragIK()->numPinnedLinks() == 0 && baseLink){
            ikPath = getCustomJointPath(body, baseLink, targetLink);
            if(ikPath){
                if(!ikPath->hasAnalyticalIK()){
                    ikPath->setBestEffortIKmode(true);
                }
                ik = ikPath;
            }
        }
    }
    
    if(!ik){
        pinDragIK = bodyItem->pinDragIK();
        pinDragIK->setBaseLink(baseLink);
        pinDragIK->setTargetLink(targetLink, kinematicsBar->isAttitudeMode());
        if(pinDragIK->initialize()){
            ik = pinDragIK;
        }
    }
    
    if(ik){

        projector.reset();

        if(kinematicsBar->isAttitudeMode()){

            if(rotationAxis >= X){
                orgAttitude = targetLink->R();
                const Matrix3& R = orgAttitude;
                axis = R.col(rotationAxis);
                const Vector3 arm = orgPointerPos - (axis.dot(orgPointerPos - orgTargetLinkPos) * axis + orgTargetLinkPos);
                rotationBaseX = arm.normalized();
                rotationBaseY = axis.cross(arm).normalized();
                Quat rotation;
                rotation.setFromTwoVectors(Vector3::UnitZ(), axis);
                projector = make_shared_aligned<SceneCylinderProjector>(
                    orgTargetLinkPos, arm.norm(), numeric_limits<float>::max(), rotation);
                dragMode = LINK_IK_ROTATION;
            }

        }

        if(!projector){
            const SgAffine3 T = event.viewMatrix().inverse().cast<SgAffine3::Scalar>();
            const Vector3 direction = -SgCamera::direction(T).cast<Vector3::Scalar>();
            projector = make_shared_aligned<ScenePlaneProjector>(direction, event.point());
            dragMode = LINK_IK_TRANSLATION;
        }

        fkTraverse.find(baseLink ? baseLink : body->rootLink(), true, true);

        if(kinematicsBar->isPenetrationBlockMode()){
            penetrationBlocker = bodyItem->createPenetrationBlocker(targetLink, true);
        } else {
            penetrationBlocker.reset();
        }
    }

    self->notifyUpdate();

}


void SceneBodyImpl::dragIK(const SceneViewEvent& event)
{
    Vector3 p;
    if(projector->project(event, p)){
        Vector3 newPos;
        Matrix3 newR;
        if(dragMode == LINK_IK_TRANSLATION){
            newPos = (orgTargetLinkPos + (Vector3(p.x(), p.y(), p.z()) - orgPointerPos));
            newR = targetLink->R();

            if(penetrationBlocker){
                penetrationBlocker->adjust(newPos, newR, newPos - targetLink->p());
            }
            
        } else if(dragMode == LINK_IK_ROTATION){
            const Vector3 r = Vector3(p.x(), p.y(), p.z()) - orgTargetLinkPos;
            const double angle = atan2(r.dot(rotationBaseY), r.dot(rotationBaseX));
            newR = AngleAxisd(angle, axis) * orgAttitude;
            newPos = orgTargetLinkPos;
        }
        if(ik->calcInverseKinematics(newPos, newR) || true /* Best effort */){
            fkTraverse.calcForwardKinematics();
            bodyItem->notifyKinematicStateChange(true);
        }
    }
}


void SceneBodyImpl::startFK(const SceneViewEvent& event)
{
    if(TRACE_FUNCTIONS){
        cout << "SceneBodyImpl::startFK()" << endl;
    }
    
    orgJointPosition = targetLink->q();
    orgPointerPos << event.point().x(), event.point().y(), event.point().z();
    
    if(targetLink->isRotationalJoint()){

        const Vector3 axis = targetLink->R() * targetLink->a();
        const Vector3 arm = orgPointerPos - (axis.dot(orgPointerPos - targetLink->p()) * axis + targetLink->p());

        if(arm.norm() > 1.0e-6){
        
            rotationBaseX = arm.normalized();
            rotationBaseY = axis.cross(rotationBaseX);
            const SgAffine3 T = event.viewMatrix().inverse().cast<SgAffine3::Scalar>();
            SgVector3 eyeDirection = SgCamera::direction(T);
            eyeDirection.normalize();
            Vector3 eyeDirection0(eyeDirection.x(), eyeDirection.y(), eyeDirection.z());
   
            if(fabs((axis.dot(eyeDirection0))) > 0.4){
                Vector3 point = event.point();
                projector = make_shared_aligned<ScenePlaneProjector>(axis, point);
            } else {
                Quat rotation;
                rotation.setFromTwoVectors(Vector3::UnitZ(), axis);
                projector = make_shared_aligned<SceneCylinderProjector>(
                    targetLink->p(), arm.norm(), numeric_limits<float>::max(), rotation);
            }
            dragMode = LINK_FK_ROTATION;
        }
        
    } else if(targetLink->isSlideJoint()){
        const SgAffine3 T = event.viewMatrix().inverse().cast<SgAffine3::Scalar>();
        SgVector3 eye = T.translation();
        SgVector3 cen = SgCamera::direction(T) + eye;
        Vector3 z = (Vector3(eye.x(), eye.y(), eye.z()) - Vector3(cen.x(), cen.y(), cen.z())).normalized();
        const Vector3 axis = targetLink->R() * targetLink->d();
        if(fabs(axis.dot(z)) < 0.95){
            const Vector3 n = (AngleAxis(M_PI / 2.0, z) * axis).cross(axis).normalized();
            projector = make_shared_aligned<ScenePlaneProjector>(n, event.point());
            dragMode = LINK_FK_TRANSLATION;
        }
    }
}


void SceneBodyImpl::dragFKRotation(const SceneViewEvent& event)
{
    Vector3 pos0;
    if(projector->project(event, pos0)){
        const Vector3 p = Vector3(pos0.x(), pos0.y(), pos0.z()) - targetLink->p();
        const double x = p.dot(rotationBaseX);
        const double y = p.dot(rotationBaseY);
        const double angle = atan2(y, x);
        targetLink->q() = orgJointPosition + angle;
        bodyItem->notifyKinematicStateChange(true);
    }
}


void SceneBodyImpl::dragFKTranslation(const SceneViewEvent& event)
{
    Vector3 p;
    if(projector->project(event, p)){
        const Vector3 axis = targetLink->R() * targetLink->d();
        Vector3 diff = (Vector3(p.x(), p.y(), p.z()) - orgPointerPos);
        double offset = axis.dot(diff);
        targetLink->q() = orgJointPosition + offset;
        bodyItem->notifyKinematicStateChange(true);
    }
}


void SceneBodyImpl::startZmpTranslation(const SceneViewEvent& event)
{
    orgPointerPos = event.point();
    projector = make_shared_aligned<ScenePlaneProjector>(Vector3::UnitZ(), event.point());
    orgZmpPos = bodyItem->zmp();
    dragMode = ZMP_TRANSLATION;
}


void SceneBodyImpl::dragZmpTranslation(const SceneViewEvent& event)
{
    Vector3 p;
    if(projector->project(event, p)){
        bodyItem->setZmp(orgZmpPos + (Vector3(p.x(), p.y(), p.z()) - orgPointerPos));
        bodyItem->notifyKinematicStateChange(true);
    }
}
