/*!
  @file
  @author Shin'ichiro Nakaoka
*/

#include "SimulatorItem.h"
#include "WorldItem.h"
#include "ControllerItem.h"
#include "BodyMotionItem.h"
#include "BodyMotionEngine.h"
#include <cnoid/AppUtil>
#include <cnoid/ExtensionManager>
#include <cnoid/TimeBar>
#include <cnoid/ItemTreeView>
#include <cnoid/MessageView>
#include <cnoid/LazyCaller>
#include <cnoid/Archive>
#include <cnoid/MultiDeviceStateSeq>
#include <cnoid/Deque2D>
#include <cnoid/ConnectionSet>
#include <cnoid/Sleep>
#include <QThread>
#include <QMutex>
#include <boost/bind.hpp>
#include <boost/dynamic_bitset.hpp>

#if QT_VERSION >= 0x040700
#include <QElapsedTimer>
#else
#include <QTime>
typedef QTime QElapsedTimer;
#endif
#include <iostream>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace cnoid;

namespace {
    typedef Deque2D<SE3, Eigen::aligned_allocator<SE3> > MultiSE3Deque;

    class SimulatedMotionEngine;
    typedef boost::shared_ptr<SimulatedMotionEngine> SimulatedMotionEnginePtr;
}

namespace cnoid {

    class SimulationBodyImpl
    {
    public:
        BodyPtr body;
        BodyItemPtr bodyItem;
        ControllerItemPtr controller;
        double frameRate;
        SimulatorItemImpl* simImpl;
        ConnectionSet deviceStateConnections;
        boost::dynamic_bitset<> deviceStateChangeFlag;

        Deque2D<double> jointPosBuf;
        MultiSE3Deque linkPosBuf;
        Deque2D<DeviceStatePtr> deviceStateBuf;
        vector<DeviceStatePtr> prevFlushedDeviceStateInDirectMode;
        MultiDeviceStateSeqPtr deviceStateResult;
        
        MultiValueSeqPtr jointPosResult;
        MultiSE3SeqPtr linkPosResult;
        MultiSE3SeqItemPtr linkPosResultItem;

        SimulationBodyImpl(const BodyPtr& body);
        bool findControllerItem(Item* item, Item*& foundItem);
        bool initialize(SimulatorItemImpl* simImpl, BodyItem* bodyItem);
        bool initialize(SimulatorItemImpl* simImpl, ControllerItem* controllerItem);
        void setupController(SimulatorItemImpl* simImpl, ControllerItem* controllerItem);
        void setupResultMotion(
            SimulatorItemImpl* simImpl, Item* ownerItem, const string& simulatedMotionName);
        void setInitialStateOfBodyMotion(const BodyMotionPtr& bodyMotion);
        void onDeviceStateChanged(int deviceIndex);
        void storeResult();
        void flushResult();
    };


    class SimulatorItemImpl : QThread
    {
    public:
        SimulatorItemImpl(SimulatorItem* self);
        SimulatorItemImpl(SimulatorItem* self, const SimulatorItemImpl& org);
        ~SimulatorItemImpl();
            
        SimulatorItem* self;

        list<SimulationBodyPtr> allSimBodies;
        vector<SimulationBodyPtr> activeSimBodies;
        vector<ControllerItem*> activeControllers;
        vector<SimulationBodyImpl*> simBodyImplsToFlush;
        bool needToUpdateSimBodyLists;
        bool hasActiveFreeBodies;

        int currentFrame;
        double worldFrameRate;
        int frameAtLastBufferWriting;
        int flushTimingCounter;
        LazyCaller flushCaller;

        Selection recordingMode;
        bool isDirectMode;
        double specifiedRecordingTimeLength;
        bool isActiveControlPeriodOnlyMode;
        int maxFrame;
        bool isAllLinkPositionOutputMode;
        bool isDeviceStateRecordingEnabled;
        bool isDoingSimulationLoop;
        bool stopRequested;
        bool isRealtimeSyncMode;
        TimeBar* timeBar;
        int fillLevelId;
        QMutex mutex;
        double actualSimulationTime;
        double finishTime;
        ostream& os;
        bool doReset;

        bool isWaitingForSimulationToStop;
        boost::signal<void()> sigSimulationFinished_;

        SimulatedMotionEnginePtr motionEngine;

        boost::signals::connection aboutToQuitConnection;
 
        void selectMotionItems();
        ControllerItem* createBodyMotionController(BodyItem* bodyItem, BodyMotionItem* bodyMotionItem);
        void forAllColdetLinkPairs(
            boost::function<void(const BodyPtr& body1, Link* link1, const BodyPtr& body2, Link* link2)> callback);
        void findTargetItems(Item* item, bool isUnderBodyItem, ItemList<Item>& out_targetItems);
        bool startSimulation(bool doReset);
        virtual void run();
        void updateSimBodyLists();
        bool stepSimulationMain();
        void flushResult();
        void stopSimulation(bool doSync);
        void onSimulationLoopStopped();
        bool onRealtimeSyncChanged(bool on);
        bool onAllLinkPositionOutputModeChanged(bool on);
        bool setSpecifiedRecordingTimeLength(double length);

        static TimeSyncItemEnginePtr getSimulatedMotionEngine(Item* sourceItem);
    };

}

namespace {

    class ControllerTarget : public ControllerItem::Target
    {
    public:
        BodyPtr body_;
        double worldTimeStep_;
        virtual BodyPtr body() const { return body_; }
        virtual double worldTimeStep() const { return worldTimeStep_; }
    };

    class SimulatedMotionEngine : public TimeSyncItemEngine
    {
      public:
        vector<BodyMotionEnginePtr> engines;

        void clear() {
            engines.clear();
        }

        void addMotionItem(BodyMotionItem* motionItem)
        {
            BodyItem* bodyItem = motionItem->findOwnerItem<BodyItem>();
            if(bodyItem){
                engines.push_back(make_shared<BodyMotionEngine>(bodyItem, motionItem));
            }
        }

        virtual bool onTimeChanged(double time) {
            bool processed = false;
            for(size_t i=0; i < engines.size(); ++i){
                processed |= engines[i]->onTimeChanged(time);
            }
            return processed;
        }

        void store(Archive& archive) {
            ListingPtr idseq = new Listing();
            idseq->setFlowStyle(true);
            for(size_t i=0; i < engines.size(); ++i){
                BodyMotionEnginePtr engine = engines[i];
                int id = archive.getItemId(engine->motionItem());
                if(id >= 0){
                    idseq->append(id);
                }
            }
            if(!idseq->empty()){
                archive.insert("motionItems", idseq);
            }
        }

        void restore(const Archive& archive) {
            archive.addPostProcess(bind(&SimulatedMotionEngine::restoreMotionItems, this, ref(archive)));
        }

        void restoreMotionItems(const Archive& archive) {
            clear();
            const Listing& idseq = *archive.findListing("motionItems");
            if(idseq.isValid()){
                for(int i=0; i < idseq.size(); ++i){
                    int id = idseq[i].toInt();
                    if(id >= 0){
                        BodyMotionItem* motionItem = dynamic_cast<BodyMotionItem*>(archive.findItem(id));
                        if(motionItem){
                            addMotionItem(motionItem);
                        }
                    }
                }
            }
        }
    };
}


void SimulatorItem::initialize(ExtensionManager* ext)
{
    ext->timeSyncItemEngineManger().addEngineFactory(SimulatorItemImpl::getSimulatedMotionEngine);
}



SimulationBody::SimulationBody(BodyPtr body)
{
    impl = new SimulationBodyImpl(body);
}


SimulationBodyImpl::SimulationBodyImpl(const BodyPtr& body)
 : body(body)
{

}


SimulationBody::~SimulationBody()
{
    delete impl;
}


BodyItem* SimulationBody::bodyItem() const
{
    return impl->bodyItem.get();
}


const BodyPtr& SimulationBody::body() const
{
    return impl->body;
}


ControllerItem* SimulationBody::controller() const
{
    return impl->controller.get();
}


bool SimulationBodyImpl::findControllerItem(Item* item, Item*& foundItem)
{
    Item* typeMatchItem = 0;
    
    while(item){
        if(dynamic_cast<ControllerItem*>(item)){
            typeMatchItem = item;
        } else if(dynamic_cast<BodyMotionItem*>(item)){
            typeMatchItem = item;
        }
        if(typeMatchItem){
            if(ItemTreeView::instance()->isItemChecked(typeMatchItem)){
                foundItem = typeMatchItem;
                return true;
            }
            if(!foundItem){
                foundItem = typeMatchItem;
            }
        }
        if(item->childItem()){
            if(findControllerItem(item->childItem(), foundItem)){
                return true;
            }
        }
        item = item->nextItem();
    }

    return false;
}


bool SimulationBodyImpl::initialize(SimulatorItemImpl* simImpl, BodyItem* bodyItem)
{
    this->simImpl = simImpl;
    this->bodyItem = bodyItem;

    // needed ?
    //body->clearExternalForces();
    //body->calcForwardKinematics(true, true);

    deviceStateConnections.disconnect();

    if(body->isStaticModel()){
        return true;
    }

    Item* targetItem = 0;
    findControllerItem(bodyItem->childItem(), targetItem);
    ControllerItem* controllerItem = 0;

    string simulatedMotionName = simImpl->self->name() + "-" + bodyItem->name();
    
    if(targetItem){
        controllerItem = dynamic_cast<ControllerItem*>(targetItem);
        if(controllerItem){
            setupController(simImpl, controllerItem);
            setupResultMotion(simImpl, controllerItem, simulatedMotionName);

        } else {
            BodyMotionItem* motionItem = dynamic_cast<BodyMotionItem*>(targetItem);
            const string& name = motionItem->name();
            if(motionItem && motionItem->name() != simulatedMotionName){
                controllerItem = simImpl->createBodyMotionController(bodyItem, motionItem);
                if(controllerItem){
                    if(simImpl->doReset){
                        setInitialStateOfBodyMotion(motionItem->motion());
                    }
                    setupController(simImpl, controllerItem);
                    setupResultMotion(simImpl, motionItem, simulatedMotionName);
                }
            }
        }
    }
    if(!controllerItem){
        frameRate = simImpl->worldFrameRate;
        setupResultMotion(simImpl, bodyItem, simulatedMotionName);
    }
    
    return true;
}


void SimulationBodyImpl::setupController(SimulatorItemImpl* simImpl, ControllerItem* controllerItem)
{
    frameRate = simImpl->worldFrameRate;
    controller = controllerItem;
}
    

void SimulationBodyImpl::setupResultMotion
(SimulatorItemImpl* simImpl, Item* ownerItem, const string& simulatedMotionName)
{
    BodyMotionItem* motionItem = dynamic_cast<BodyMotionItem*>(ownerItem);
    if(motionItem){
        BodyMotionItem* resultMotionItem = motionItem->findItem<BodyMotionItem>(simulatedMotionName);
        if(resultMotionItem){
            motionItem = resultMotionItem;
        } else {
            resultMotionItem = new BodyMotionItem();
            resultMotionItem->setTemporal();
            resultMotionItem->setName(simulatedMotionName);
            motionItem->addChildItem(resultMotionItem);
            motionItem = resultMotionItem;
        }
    } else {
        motionItem = ownerItem->findItem<BodyMotionItem>(simulatedMotionName);
        if(!motionItem){
            motionItem = new BodyMotionItem();
            motionItem->setTemporal();
            motionItem->setName(simulatedMotionName);
            ownerItem->addChildItem(motionItem);
        }
    }

    simImpl->motionEngine->addMotionItem(motionItem);

    const int numJoints = body->numJoints();
    const int numLinksToRecord = simImpl->isAllLinkPositionOutputMode ? body->numLinks() : 1;
    BodyMotion& motion = *motionItem->motion();
    motion.setFrameRate(frameRate);
    motion.setDimension(1, numJoints, numLinksToRecord);
    
    jointPosBuf.resizeColumn(numJoints);
    jointPosResult = motion.jointPosSeq();
    if(numJoints > 0){
        MultiValueSeq::Frame jframe0 = jointPosResult->frame(0);
        for(int i=0; i < numJoints; ++i){
            jframe0[i] = body->joint(i)->q();
        }
    }

    linkPosBuf.resizeColumn(numLinksToRecord);
    linkPosResultItem = motionItem->linkPosSeqItem();
    linkPosResult = motion.linkPosSeq();
    MultiSE3Seq::Frame lframe0 = linkPosResult->frame(0);
    for(int i=0; i < numLinksToRecord; ++i){
        Link* link = body->link(i);
        lframe0[i].set(link->p(), link->R());
    }
    
    const DeviceList<>& devices = body->devices();

    if(devices.empty() || !simImpl->isDeviceStateRecordingEnabled){
        deviceStateChangeFlag.clear();
        deviceStateBuf.clear();
        clearMultiDeviceStateSeq(motion);
        prevFlushedDeviceStateInDirectMode.clear();
    } else {
        deviceStateChangeFlag.resize(devices.size());
        deviceStateChangeFlag.reset();
        deviceStateBuf.resize(1, devices.size());
        prevFlushedDeviceStateInDirectMode.resize(devices.size());
        Deque2D<DeviceStatePtr>::Row dstate = deviceStateBuf.last();

        for(size_t i=0; i < devices.size(); ++i){
            DeviceState* s = devices[i]->cloneState();
            dstate[i] = s;
            prevFlushedDeviceStateInDirectMode[i] = s;

            deviceStateConnections.add(
                devices[i]->sigStateChanged().connect(
                    bind(&SimulationBodyImpl::onDeviceStateChanged, this, i)));
        }
        if(!simImpl->isDirectMode){
            deviceStateResult = getOrCreateMultiDeviceStateSeq(motion);
            deviceStateResult->setNumParts(devices.size());
            MultiDeviceStateSeq::Row result0 = deviceStateResult->frame(0);
            for(size_t i=0; i < dstate.size(); ++i){
                result0[i] = dstate[i];
            }
        }
    }

    /*
    if(simImpl->self->recordingMode() != SimulatorItem::DIRECT){
        ItemTreeView::instance()->selectItem(motionItem);
    }
    */
}


void SimulationBodyImpl::setInitialStateOfBodyMotion(const BodyMotionPtr& bodyMotion)
{
    bool updated = false;
    
    MultiSE3SeqPtr lseq = bodyMotion->linkPosSeq();
    if(lseq->numParts() > 0 && lseq->numFrames() > 0){
        SE3& p = lseq->at(0, 0);
        Link* rootLink = body->rootLink();
        rootLink->p() = p.translation();
        rootLink->R() = p.rotation().toRotationMatrix();
        updated = true;
    }
    MultiValueSeqPtr jseq = bodyMotion->jointPosSeq();
    if(jseq->numFrames() > 0){
        MultiValueSeq::Frame jframe0 = jseq->frame(0);
        int n = std::min(jframe0.size(), body->numJoints());
        for(int i=0; i < n; ++i){
            body->joint(i)->q() = jframe0[i];
        }
        updated = true;
    }
    if(updated){
        body->calcForwardKinematics();
    }
}


// For a controller which is not associated with a body
bool SimulationBodyImpl::initialize(SimulatorItemImpl* simImpl, ControllerItem* controllerItem)
{
    this->simImpl = simImpl;
    this->controller = controllerItem;
    frameRate = simImpl->worldFrameRate;
    linkPosBuf.resizeColumn(0);
    return true;
}


void SimulationBodyImpl::onDeviceStateChanged(int deviceIndex)
{
    deviceStateChangeFlag.set(deviceIndex);
}


void SimulationBody::storeResult()
{
    impl->storeResult();
}


void SimulationBodyImpl::storeResult()
{
    if(jointPosBuf.colSize() > 0){
        Deque2D<double>::Row q = jointPosBuf.append();
        for(int i=0; i < q.size() ; ++i){
            q[i] = body->joint(i)->q();
        }
    }
    MultiSE3Deque::Row pos = linkPosBuf.append();
    for(int i=0; i < linkPosBuf.colSize(); ++i){
        Link* link = body->link(i);
        pos[i].set(link->p(), link->R());
    }

    if(!deviceStateBuf.empty()){
        Deque2D<DeviceStatePtr>::Row current = deviceStateBuf.append();
        Deque2D<DeviceStatePtr>::Row old = deviceStateBuf[deviceStateBuf.rowSize() - 2];
        const DeviceList<>& devices = body->devices();
        for(size_t i=0; i < devices.size(); ++i){
            if(deviceStateChangeFlag[i]){
                current[i] = devices[i]->cloneState();
            } else {
                current[i] = old[i];
            }
        }
    }
}


bool SimulationBody::flushResult()
{
    if(!impl->linkPosResult || impl->linkPosBuf.rowSize() == 0){
        return false;
    }
    impl->flushResult();
    return true;
}


void SimulationBodyImpl::flushResult()
{
    const int frame = linkPosResult->numFrames();
    const int numBufFrames = linkPosBuf.rowSize();
    const int ndevices = deviceStateBuf.colSize();

    if(simImpl->isDirectMode){
        const BodyPtr& orgBody = bodyItem->body();
        const int n = body->numJoints();
        if(n > 0){
            Deque2D<double>::Row q = jointPosBuf.last();
            for(int i=0; i < n; ++i){
                orgBody->joint(i)->q() = q[i];
            }
        }
        for(int i=0; i < linkPosBuf.colSize(); ++i){
            SE3& pos = linkPosBuf.last()[i];
            Link* link = orgBody->link(i);
            link->p() = pos.translation();
            link->R() = pos.rotation().toRotationMatrix();
        }
        const DeviceList<>& devices = orgBody->devices();
        if(ndevices > 0){
            Deque2D<DeviceStatePtr>::Row ds = deviceStateBuf.last();
            for(size_t i=0; i < ndevices; ++i){
                const DeviceStatePtr& s = ds[i];
                if(s != prevFlushedDeviceStateInDirectMode[i]){
                    Device* device = devices.get(i);
                    device->copyStateFrom(*s);
                    device->notifyStateChange();
                    prevFlushedDeviceStateInDirectMode[i] = s;
                }
            }
            // keep the last state so that unchanged states can be shared
            deviceStateBuf.pop_front(deviceStateBuf.rowSize() - 1);
        }
        
    } else {
        if(jointPosBuf.colSize() > 0){
            jointPosResult->setNumFrames(frame + numBufFrames);
            for(int i=0; i < numBufFrames; ++i){
                Deque2D<double>::Row buf = jointPosBuf.row(i);
                std::copy(buf.begin(), buf.end(), jointPosResult->frame(frame + i).begin());
            }
        }
        linkPosResult->setNumFrames(frame + numBufFrames);
        for(int i=0; i < numBufFrames; ++i){
            MultiSE3Deque::Row buf = linkPosBuf.row(i);
            std::copy(buf.begin(), buf.end(), linkPosResult->frame(frame + i).begin());
        }
        if(ndevices > 0){
            deviceStateResult->setNumFrames(frame + numBufFrames);
            for(int i=0; i < numBufFrames; ++i){
                Deque2D<DeviceStatePtr>::Row buf = deviceStateBuf.row(i);
                std::copy(buf.begin(), buf.end(), deviceStateResult->frame(frame + i).begin());
            }
            // keep the last state so that unchanged states can be shared
            deviceStateBuf.pop_front(deviceStateBuf.rowSize() - 1);
        }
    }
    jointPosBuf.resizeRow(0);
    linkPosBuf.resizeRow(0);
}


SimulatorItem::SimulatorItem()
{
    impl = new SimulatorItemImpl(this);
}


SimulatorItem::SimulatorItem(const SimulatorItem& org)
    : Item(org)
{
    impl = new SimulatorItemImpl(this);
    impl->isRealtimeSyncMode = org.impl->isRealtimeSyncMode;
    impl->isAllLinkPositionOutputMode = org.impl->isAllLinkPositionOutputMode;
    impl->isDeviceStateRecordingEnabled = org.impl->isDeviceStateRecordingEnabled;
    impl->recordingMode = org.impl->recordingMode;
    impl->isActiveControlPeriodOnlyMode = org.impl->isActiveControlPeriodOnlyMode;
}


SimulatorItemImpl::SimulatorItemImpl(SimulatorItem* self)
    : self(self),
      os(MessageView::mainInstance()->cout()),
      flushCaller(bind(&SimulatorItemImpl::flushResult, this)),
      recordingMode(SimulatorItem::N_RECORDING_MODES, CNOID_GETTEXT_DOMAIN_NAME)
{
    timeBar = TimeBar::instance();
    isDoingSimulationLoop = false;
    isAllLinkPositionOutputMode = false;
    isDeviceStateRecordingEnabled = true;
    isRealtimeSyncMode = false;

    recordingMode.setSymbol(SimulatorItem::UNLIMITED,     N_("Unlimited"));
    recordingMode.setSymbol(SimulatorItem::SPECIFIC_TIME, N_("Specific time length"));
    recordingMode.setSymbol(SimulatorItem::TIMEBAR_RANGE, N_("TimeBar range"));
    recordingMode.setSymbol(SimulatorItem::DIRECT,        N_("Direct"));

    recordingMode.select(SimulatorItem::TIMEBAR_RANGE);

    specifiedRecordingTimeLength = 60.0;
    
    isActiveControlPeriodOnlyMode = true;

    motionEngine = make_shared<SimulatedMotionEngine>();
}


SimulatorItem::~SimulatorItem()
{
    impl->stopSimulation(true);
    delete impl;
}


SimulatorItemImpl::~SimulatorItemImpl()
{
    aboutToQuitConnection.disconnect();
}


void SimulatorItem::onDisconnectedFromRoot()
{
    impl->stopSimulation(true);
}


void SimulatorItem::setAllLinkPositionOutputMode(bool on)
{
    impl->isAllLinkPositionOutputMode = on;
}


bool SimulatorItem::isAllLinkPositionOutputMode()
{
    return impl->isAllLinkPositionOutputMode;
}


double SimulatorItem::worldTimeStep()
{
    return TimeBar::instance()->timeStep();
}


SimulatorItem::RecordingMode SimulatorItem::recordingMode() const
{
    return static_cast<SimulatorItem::RecordingMode>(impl->recordingMode.selectedIndex());
}


bool SimulatorItemImpl::setSpecifiedRecordingTimeLength(double length)
{
    specifiedRecordingTimeLength = length;
	return true;
}


void SimulatorItem::selectMotionItems()
{
    impl->selectMotionItems();
}


void SimulatorItemImpl::selectMotionItems()
{
    ItemTreeView* itemTreeView = ItemTreeView::instance();

    for(list<SimulationBodyPtr>::iterator p = allSimBodies.begin(); p != allSimBodies.end(); ++p){
        MultiSE3SeqItem* linkPosResultItem = (*p)->impl->linkPosResultItem.get();
        if(linkPosResultItem){
            BodyMotionItem* bodyMotionItem = dynamic_cast<BodyMotionItem*>(linkPosResultItem->parentItem());
            if(bodyMotionItem){
                itemTreeView->selectItem(bodyMotionItem);
            } else {
                itemTreeView->selectItem(linkPosResultItem);
            }
        }
    }
}


ControllerItem* SimulatorItem::createBodyMotionController(BodyItem* bodyItem, BodyMotionItem* bodyMotionItem)
{
    return 0;
}


/*
  Extract body items and controller items which are not associated with (not under) a body item
*/
void SimulatorItemImpl::findTargetItems(Item* item, bool isUnderBodyItem, ItemList<Item>& out_targetItems)
{
    if(dynamic_cast<BodyItem*>(item)){
        out_targetItems.push_back(item);
        isUnderBodyItem = true;
    } else if(!isUnderBodyItem){
        if(dynamic_cast<ControllerItem*>(item)){
            out_targetItems.push_back(item);
        }
    }
    for(Item* childItem = item->childItem(); childItem; childItem = childItem->nextItem()){
        findTargetItems(childItem, isUnderBodyItem, out_targetItems);
    }
}


bool SimulatorItem::startSimulation(bool doReset)
{
    return impl->startSimulation(doReset);
}


bool SimulatorItemImpl::startSimulation(bool doReset)
{
    this->doReset = doReset;
    
    stopSimulation(true);

    WorldItem* worldItem = self->findOwnerItem<WorldItem>();
    if(!worldItem){
        os << (fmt(_("%1% must be in a WorldItem to do simulation.")) % self->name()) << endl;
        return false;
    }

    ItemList<Item> targetItems;
    findTargetItems(worldItem, false, targetItems);
    if(targetItems.empty()){
        return false;
    }

    isDirectMode = (self->recordingMode() == SimulatorItem::DIRECT);
    worldFrameRate = 1.0 / self->worldTimeStep();

    allSimBodies.clear();
    activeSimBodies.clear();
    motionEngine->clear();
    needToUpdateSimBodyLists = true;
    vector<SimulationBodyPtr> simBodiesWithBody;
    
    for(size_t i=0; i < targetItems.size(); ++i){

        if(BodyItem* bodyItem = dynamic_cast<BodyItem*>(targetItems.get(i))){
            if(doReset){
                bodyItem->restoreInitialState();
            }
            SimulationBodyPtr simBody = self->createSimulationBody(*bodyItem->body());
            if(simBody->body()){
                if(simBody->impl->initialize(this, bodyItem)){
                    allSimBodies.push_back(simBody);
                    simBodiesWithBody.push_back(simBody);
                }
            }
        } else {
            // ControllerItem which is not associated with a body
            ControllerItem* controller = dynamic_cast<ControllerItem*>(targetItems.get(i));
            if(controller){
                SimulationBodyPtr simBody = make_shared<SimulationBody>(BodyPtr());
                if(simBody->impl->initialize(this, controller)){
                    allSimBodies.push_back(simBody);
                }
            }
        }
    }
    
    bool result = self->initializeSimulation(simBodiesWithBody);

    if(result){

        ControllerTarget target;
        target.worldTimeStep_ = self->worldTimeStep();

        for(list<SimulationBodyPtr>::iterator p = allSimBodies.begin(); p != allSimBodies.end(); ++p){
            SimulationBodyImpl* simBodyImpl = (*p)->impl;
            ControllerItemPtr controller = simBodyImpl->controller;
            if(controller){
                const BodyPtr& body = simBodyImpl->body;
                bool ready = false;
                controller->setSimulatorItem(self);
                if(body){
                    target.body_ = body;
                    ready = controller->start(target);
                    if(!ready){
                        os << (fmt(_("%1% for %2% failed to initialize."))
                               % controller->name() % simBodyImpl->bodyItem->name()) << endl;
                    }
                } else {
                    target.body_.reset(); // pass null body
                    ready = controller->start(target);
                    if(!ready){
                        os << (fmt(_("%1% failed to initialize."))
                               % controller->name()) << endl;
                    }
                }
                if(!ready){
                    controller->setSimulatorItem(0);
                    string message = controller->getMessage();
                    if(!message.empty()){
                        os << message << endl;
                    }
                    simBodyImpl->controller = 0;
                }
            }
        }

        currentFrame = 0;
        frameAtLastBufferWriting = 0;
        flushTimingCounter = 0;

        if(recordingMode.is(SimulatorItem::SPECIFIC_TIME)){
            maxFrame = specifiedRecordingTimeLength / self->worldTimeStep();
        } else if(recordingMode.is(SimulatorItem::TIMEBAR_RANGE)){
            maxFrame = TimeBar::instance()->maxTime() / self->worldTimeStep();
        } else {
            maxFrame = std::numeric_limits<int>::max();
        }

        isDoingSimulationLoop = true;
        isWaitingForSimulationToStop = false;
        stopRequested = false;

        aboutToQuitConnection.disconnect();
        aboutToQuitConnection = cnoid::sigAboutToQuit().connect(bind(&SimulatorItemImpl::stopSimulation, this, true));

        start();
        os << (fmt(_("Simulation by %1% has started.")) % self->name()) << endl;

        if(!recordingMode.is(SimulatorItem::DIRECT)){
            fillLevelId = timeBar->startFillLevelUpdate();
            if(!timeBar->isDoingPlayback()){
                timeBar->setTime(0.0);
                timeBar->startPlayback();
            }
        }
    }

    return result;
}


ControllerItem* SimulatorItemImpl::createBodyMotionController(BodyItem* bodyItem, BodyMotionItem* bodyMotionItem)
{
    return self->createBodyMotionController(bodyItem, bodyMotionItem);
}


void SimulatorItem::forAllColdetLinkPairs
(boost::function<void(const BodyPtr& body1, Link* link1, const BodyPtr& body2, Link* link2)> callback)
{
    return impl->forAllColdetLinkPairs(callback);
}


void SimulatorItemImpl::forAllColdetLinkPairs
(boost::function<void(const BodyPtr& body1, Link* link1, const BodyPtr& body2, Link* link2)> callback)
{
    typedef std::list<SimulationBodyPtr>::iterator Iter;

    for(Iter b1 = allSimBodies.begin(); b1 != allSimBodies.end(); ++b1){
        const BodyPtr& body1 = (*b1)->body();
        if(body1){
            Iter b2 = b1;
            ++b2;
            for( ; b2 != allSimBodies.end(); ++b2){
                const BodyPtr& body2 = (*b2)->body();
                if(body2){
                    for(int l1 = 0; l1 < body1->numLinks(); ++l1){
                        Link* link1 = body1->link(l1);
                        if(link1->coldetModel()){
                            bool isLink1Static = body1->isStaticModel() || (!link1->parent() && (link1->isFixedJoint()));
                            for(int l2 = 0; l2 < body2->numLinks(); ++l2){
                                Link* link2 = body2->link(l2);
                                if(link2->coldetModel()){
                                    if(link1 != link2){
                                        bool isLink2Static = body2->isStaticModel() || (!link2->parent() && (link2->isFixedJoint()));
                                        if(!isLink1Static || !isLink2Static){
                                            callback(body1, link1, body2, link2);
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}


void SimulatorItem::initializeSimulationThread()
{

}


// Simulation loop
void SimulatorItemImpl::run()
{
    self->initializeSimulationThread();

    QElapsedTimer timer;
    timer.start();

    int frame = 0;

    if(isRealtimeSyncMode){
        double dtms = self->worldTimeStep() * 1000.0;
        double tms0 = 0.0;
        while(true){
            if(!stepSimulationMain() || stopRequested || frame >= maxFrame){
                break;
            }
            double tms = dtms * frame;
            if(tms - tms0 >= 10.0){
                tms0 += 10;
                double d = tms - timer.elapsed();
                if(d > 10.0){
                    QThread::msleep(d - 1.0);
                }
            }
            frame++;
        }
    } else {
        while(true){
            if(!stepSimulationMain() || stopRequested || frame++ >= maxFrame){
                break;
            }
        }
    }

    actualSimulationTime = (timer.elapsed() / 1000.0);
    finishTime = frame / worldFrameRate;

    isDoingSimulationLoop = false;

    if(!isWaitingForSimulationToStop){
        callLater(bind(&SimulatorItemImpl::onSimulationLoopStopped, this));
    }
}


void SimulatorItemImpl::updateSimBodyLists()
{
    activeSimBodies.clear();
    activeControllers.clear();
    hasActiveFreeBodies = false;
    
    for(list<SimulationBodyPtr>::iterator p = allSimBodies.begin(); p != allSimBodies.end(); ++p){
        SimulationBodyPtr& simBody = *p;
        SimulationBodyImpl* simBodyImpl = simBody->impl;
        ControllerItem* controller = simBodyImpl->controller.get();
        if(controller){
            activeControllers.push_back(controller);
            if(simBodyImpl->body && !simBodyImpl->body->isStaticModel()){
                activeSimBodies.push_back(simBody);
            }
        } else if(simBodyImpl->body && !simBodyImpl->body->isStaticModel()){
            activeSimBodies.push_back(simBody);
            hasActiveFreeBodies = true;
        }
    }

    needToUpdateSimBodyLists = false;
}


bool SimulatorItemImpl::stepSimulationMain()
{
    currentFrame++;

    if(needToUpdateSimBodyLists){
        updateSimBodyLists();
    }
    
    bool doContinue = hasActiveFreeBodies || !isActiveControlPeriodOnlyMode;

    for(size_t i=0; i < activeControllers.size(); ++i){
        ControllerItem* controller = activeControllers[i];
        if(controller->isImmediateMode()){
            controller->input();
            doContinue |= controller->control();
            controller->output();
        } else {
            doContinue |= controller->control();
            // The status before the next forward dynamics
            // is used for the input of the next control 
            controller->input();
        }
    }

    if(doContinue){

        self->stepSimulation(activeSimBodies);

        mutex.lock();
        
        for(size_t i=0; i < activeSimBodies.size(); ++i){
            activeSimBodies[i]->storeResult();
        }
        frameAtLastBufferWriting = currentFrame;

        if(++flushTimingCounter == 10){
            //callLater(bind(&SimulatorItemImpl::flushResult, this));
            flushCaller.request();
            flushTimingCounter = 0;
        }

        mutex.unlock();

        for(size_t i=0; i < activeControllers.size(); ++i){
            ControllerItem* controller = activeControllers[i];
            if(!controller->isImmediateMode()){
                // the output of the previous control is used for the next forward dynamics
                controller->output(); 
            }
        }
    }

    return doContinue;
}


void SimulatorItemImpl::flushResult()
{
    simBodyImplsToFlush.clear();

    mutex.lock();
    
    for(list<SimulationBodyPtr>::iterator p = allSimBodies.begin(); p != allSimBodies.end(); ++p){
        SimulationBodyPtr& simBody = *p;
        if(simBody->flushResult()){
            simBodyImplsToFlush.push_back(simBody->impl);
        }
    }

    int frame = frameAtLastBufferWriting;
    
    mutex.unlock();

    if(isDirectMode){
        for(size_t i=0; i < simBodyImplsToFlush.size(); ++i){
            simBodyImplsToFlush[i]->bodyItem->notifyKinematicStateChange(!isAllLinkPositionOutputMode);
        }
    } else {
        for(size_t i=0; i < simBodyImplsToFlush.size(); ++i){
            //simBodyImplsToFlush[i]->linkPosResultItem->notifyUpdate();
            
            // Should this code replaced with a following kind of cuntion?
            // BodyMotionEngine::instance()->update();
        }
    }

    double fillLevel = frame / worldFrameRate;

    timeBar->updateFillLevel(fillLevelId, fillLevel);
}


void SimulatorItem::stopSimulation()
{
    impl->stopSimulation(false);
}


void SimulatorItemImpl::stopSimulation(bool doSync)
{
    if(isDoingSimulationLoop){
        if(doSync){
            isWaitingForSimulationToStop = true;
        }
        stopRequested = true;
        
        if(doSync){
            wait();
            isWaitingForSimulationToStop = false;
            onSimulationLoopStopped();
        }
    }
    aboutToQuitConnection.disconnect();
}


void SimulatorItem::finalizeSimulation()
{

}


void SimulatorItemImpl::onSimulationLoopStopped()
{
    for(list<SimulationBodyPtr>::iterator p = allSimBodies.begin(); p != allSimBodies.end(); ++p){
        ControllerItemPtr& controller = (*p)->impl->controller;
        if(controller){
            controller->stop();
            controller->setSimulatorItem(0);
        }
    }
    self->finalizeSimulation();

    flushResult();

    timeBar->stopFillLevelUpdate(fillLevelId);

    os << (fmt(_("Simulation by %1% has finished at %2% [s].\n")) % self->name() % finishTime);
    os << (fmt(_(" Computation time is %1% [s], computation time / simulation time = %2%."))
           % actualSimulationTime % (actualSimulationTime / finishTime))
       << endl;

    sigSimulationFinished_();

    allSimBodies.clear();
    activeSimBodies.clear();
}


bool SimulatorItem::isRunning() const
{
    return impl->isDoingSimulationLoop;
}


int SimulatorItem::currentFrame() const
{
    return impl->currentFrame;
}


SignalProxy< boost::signal<void()> > SimulatorItem::sigSimulationFinished()
{
    return impl->sigSimulationFinished_;
}


TimeSyncItemEnginePtr SimulatorItemImpl::getSimulatedMotionEngine(Item* sourceItem)
{
    SimulatorItem* simulatorItem = dynamic_cast<SimulatorItem*>(sourceItem);
    return simulatorItem ? simulatorItem->impl->motionEngine : TimeSyncItemEnginePtr();
}


bool SimulatorItemImpl::onRealtimeSyncChanged(bool on)
{
    isRealtimeSyncMode = on;
    return true;
}


bool SimulatorItemImpl::onAllLinkPositionOutputModeChanged(bool on)
{
    self->setAllLinkPositionOutputMode(on);
    return (isAllLinkPositionOutputMode == on);
}


void SimulatorItem::doPutProperties(PutPropertyFunction& putProperty)
{
    putProperty(_("Sync with realtime"), impl->isRealtimeSyncMode,
                bind(&SimulatorItemImpl::onRealtimeSyncChanged, impl, _1));
    putProperty(_("Recording mode"), impl->recordingMode,
                bind((bool(Selection::*)(int))&Selection::select, &impl->recordingMode, _1));
    putProperty(_("Active control period only"), impl->isActiveControlPeriodOnlyMode,
                changeProperty(impl->isActiveControlPeriodOnlyMode));
    putProperty(_("Time length"), impl->specifiedRecordingTimeLength,
                bind(&SimulatorItemImpl::setSpecifiedRecordingTimeLength, impl, _1));
    putProperty(_("All link positions"), impl->isAllLinkPositionOutputMode,
                bind(&SimulatorItemImpl::onAllLinkPositionOutputModeChanged, impl, _1));
    putProperty(_("Device state recording"), impl->isDeviceStateRecordingEnabled,
                changeProperty(impl->isDeviceStateRecordingEnabled));
}


bool SimulatorItem::store(Archive& archive)
{
    archive.write("realtimeSync", impl->isRealtimeSyncMode);
    archive.write("recordingMode", impl->recordingMode.selectedSymbol());
    archive.write("onlyActiveControlPeriod", impl->isActiveControlPeriodOnlyMode);
    archive.write("timeLength", impl->specifiedRecordingTimeLength);
    archive.write("allLinkPositionOutputMode", impl->isAllLinkPositionOutputMode);
    archive.write("deviceStateRecording", impl->isDeviceStateRecordingEnabled);
    impl->motionEngine->store(archive);
    return true;
}


bool SimulatorItem::restore(const Archive& archive)
{
    archive.read("realtimeSync", impl->isRealtimeSyncMode);
    string symbol;
    if(archive.read("recordingMode", symbol)){
        impl->recordingMode.select(symbol);
    }
    archive.read("onlyActiveControlPeriod", impl->isActiveControlPeriodOnlyMode);
    archive.read("timeLength", impl->specifiedRecordingTimeLength);
    setAllLinkPositionOutputMode(archive.get("allLinkPositionOutputMode", impl->isAllLinkPositionOutputMode));
    archive.read("deviceStateRecording", impl->isDeviceStateRecordingEnabled);
    impl->motionEngine->restore(archive);
    return true;
}
